"""
Implements the "hardware" for tracing a function.

Will perform the instructions and updates CPU registers, stack information, etc.

WARNING:
    Do NOT rely on the flags registers being correct.  There are places were flags are NOT being updated when they
    should, and the very fact that CALL instructions are skipped could cause flags to be incorrect.
"""
from copy import deepcopy
import collections
import logging
import warnings
from typing import List, Tuple, Optional, Union

import ida_frame
import ida_funcs
import ida_idp
import idaapi
import idc
import ida_struct
import ida_ua

from kordesii.utils.function_tracing import utils
from kordesii.utils.function_tracing.constants import *
from kordesii.utils.function_tracing.instruction import Instruction
from kordesii.utils.function_tracing.memory import Memory
from kordesii.utils.function_tracing.variables import VariableMap
from kordesii.utils.function_tracing.operands import Operand, OperandLite
from kordesii.utils.function_tracing.functions import FunctionSignature, FunctionArg
from kordesii.utils.function_tracing.objects import File, RegKey, Service, ObjectMap
from kordesii.utils.function_tracing.actions import ActionList
from kordesii.utils.functions import Function

logger = logging.getLogger(__name__)


class JccContext(object):
    """
    Stores information pertaining to a Jcc instruction encountered when tracing.

    When a Jcc instruction is encountered, several pieces of information inherently need to be tracked since
    we are blindly taking every branch to ensure we get all possible data at any given address.  It turns out
    we need to know the target of the Jcc instruction for the condition as emulated
    (condition_target_ea).  We also need to know the value of the branch we would NOT have taken (at least as
    best of a guess as we can make in some cases) and where that value would have been set.  In order to
    calculate the value, we need to know what kind of test instruction was used, so that mnem is tracked as well.When
    we trace our condition_target_ea branch, we need not modify the context.  Whenever we trace the alternative branch,
    we'll need to modify the context as specified.
    """

    def __init__(self):
        self.condition_target_ea = None  # The branch actually taken
        self.alt_branch_data_dst = None  # The location which was tested (typically opnd 0 of the condition test)
        self.alt_branch_data = None  # The data stored in _alt_branc_data_dst
        self.flag_opnds = {}  # Dictionary containing the operands at a particular instruction which set
        # specific flags.  Dictionary is keyed on flag registery names.

    def __deepcopy__(self, memo):
        copy = JccContext()
        copy.condition_target_ea = self.condition_target_ea
        copy.alt_branch_data_dst = self.alt_branch_data_dst
        copy.alt_branch_data = self.alt_branch_data
        copy.flag_opnds = {flag: list(operands) for flag, operands in list(self.flag_opnds.items())}
        return copy

    def update_flag_opnds(self, flags, opnds):
        """
        Set the operands which last changed the specified flags.

        :param flags: list of flags which were modified utilizing the supplied opnds
        :param opnds: list of operands (instance of Operand) at the instruction which modified the flags
        """
        for flag in flags:
            # Converting Operand classes to OperandLite classes to help speed up deepcopies.
            self.flag_opnds[flag] = [OperandLite(opnd.ip, opnd.idx, opnd.text, opnd.value) for opnd in opnds]

    def get_flag_opnds(self, flags):
        """
        Extracts all the operands of for the list of flags and reduces the set.  However, since the operands
        need to remain in order, we can't use set operations.  In all actuality, assuming our code is correct and
        the compiler isn't doing something funky, any more than 1 flag should really just be a duplicate list.

        :param flags: list of flags for which to extract operands
        :return: list of operands which were utilized in the instruction that modified the requested flags
        """
        # TODO: Is there a better way to do this?
        operands = []
        for flag in flags:
            for operand in self.flag_opnds.get(flag, []):
                if operand not in operands:
                    operands.append(operand)

        return operands

    def is_alt_branch(self, ip):
        """
        Test our IP against the branch information to determine if we are in the branch that would have been
        emulated or in the alternate branch.
        """
        return self.condition_target_ea and self.condition_target_ea != ip


class ProcessorContext(object):
    """
    Stores the context of the processor during execution.

    :param emulator: Instance of Emulator to use during emulation.
    :param registers: Instance of an initialized RegisterMap object used to store register values
        for the given architecture.
    :param str instruction_pointer: Name of the register used to point to the current instruction
        being currently executed or to-be executed.
    :param str stack_pointer: Name of the register used to hold the stack pointer.
    """

    # Must be set by inherited classes.
    OPCODES = {}  # Map of opcode mnemonics to functions that emulate them.

    # Class used to generate instructions.
    _instruction_class = Instruction

    # Cache for keeping track of instructions and their operand indexes.
    _operand_indices = {}

    def __init__(self, emulator, registers, instruction_pointer, stack_pointer, return_register):
        self.emulator = emulator
        self.registers = registers
        self.jcccontext = JccContext()
        self.memory = Memory()
        self.call_history = []  # Keeps track of function calls.
        self.executed_instructions = []  # Keeps track of the instructions that have been executed.
        self.memory_copies = collections.defaultdict(list)  # Keeps track of memory moves.
        self.bitness = utils.get_bits()
        self.byteness = self.bitness // 8
        self.variables = VariableMap(self)
        self.objects = ObjectMap(self)
        self.actions = ActionList()
        self.stdout = ""

        # Function start address of a function we are currently hooking.
        self.hooking_call = None
        self._call_depth = 0

        self._sp = stack_pointer
        self._ip = instruction_pointer
        self._ret = return_register
        self._sp_start = self.sp

    def __deepcopy__(self, memo):
        """Implementing our own deepcopy to improve speed."""
        # Create class, but avoid calling __init__()
        # so we don't trigger the unnecessary initialization of Memory and JccContext
        klass = self.__class__
        copy = klass.__new__(klass)
        memo[id(self)] = copy

        copy.emulator = self.emulator  # This is a reference, don't create a new instance.
        copy.hooking_call = self.hooking_call
        copy.registers = deepcopy(self.registers, memo)
        copy.jcccontext = deepcopy(self.jcccontext, memo)
        copy.memory = deepcopy(self.memory, memo)
        copy.variables = deepcopy(self.variables, memo)
        copy.objects = deepcopy(self.objects, memo)
        copy.actions = deepcopy(self.actions, memo)
        copy.call_history = list(self.call_history)
        copy.executed_instructions = list(self.executed_instructions)
        copy.memory_copies = self.memory_copies.copy()
        copy.stdout = self.stdout
        copy.bitness = self.bitness
        copy.byteness = self.byteness
        copy._call_depth = self._call_depth
        copy._sp = self._sp
        copy._ip = self._ip
        copy._ret = self._ret
        copy._sp_start = self._sp_start

        return copy

    @property
    def ip(self) -> int:
        """Alias for retrieving instruction pointer."""
        return self.registers[self._ip]

    @ip.setter
    def ip(self, value):
        """Alias for setting instruction pointer."""
        self.registers[self._ip] = value

    @property
    def sp(self) -> int:
        """Alias for retrieving stack pointer."""
        return self.registers[self._sp]

    @sp.setter
    def sp(self, value):
        """Alias for setting stack pointer."""
        self.registers[self._sp] = value

    @property
    def sp_diff(self) -> int:
        """
        The difference between the current stack pointer and the
        stack pointer at the beginning of the function.

        This helps with debugging since this number should match the number
        shown in the IDA disassembly.
        """
        return self._sp_start - self.sp

    # TODO: A subroutine in ARM can technically pass in larger values, in which
    #   case the value spans multiple registers r0-r3
    @property
    def ret(self) -> int:
        """Alias for retrieving the return value."""
        return self.registers[self._ret]

    @ret.setter
    def ret(self, value):
        """Alias for setting return value."""
        logger.debug("Setting 0x%X into %s", value, self._ret)
        self.registers[self._ret] = value

    @property
    def prev_instruction(self):
        """That last instruction that was executed or None if no instructions have been executed."""
        if self.executed_instructions:
            return self.executed_instructions[-1]
        else:
            return None

    def execute(self, start=None, end=None, call_depth: int = 0, max_instructions: int = None):
        """
        "Execute" the instruction at IP and store results in the context.
        The instruction pointer register will be set to the value supplied in .ip so that
        it is correct.

        :param start: instruction address to start execution (defaults to currently set ip)
        :param end: instruction to stop execution (not including)
            (defaults to only run start)
        :param call_depth: Number of function calls we are allowed to emulate into.
            When we hit our limit (depth is 0), emulation will no longer jump into function calls.
            (Defaults to not emulating into any function calls.)
            NOTE: This does not affect call hooks.
        :param max_instructions: Maximum number of instructions to execute before
            raising an RuntimeError.
            Uses max_instructions set by emulator constructor if not provided.

        :raises RuntimeError: If maximum number of instructions get hit.
        """
        if max_instructions is None:
            max_instructions = self.emulator.max_instructions

        if not start:
            start = self.ip

        # Set instruction pointer to where we are currently executing.
        self.ip = start

        # Set current call depth.
        if call_depth < 0:
            raise ValueError(f"call_depth must be a positive number.")
        self._call_depth = call_depth

        # If end is provided, recursively run execute() until ip is end.
        if end is not None:
            count = max_instructions
            while self.ip != end:
                instruction = self.instruction
                if instruction.is_terminal:
                    return  # TODO: Should we be executing the terminal instruction?
                instruction.execute()
                count -= 1
                if not count:
                    raise RuntimeError('Hit maximum number of instructions.')
            return
        else:
            self.instruction.execute()

    def _execute_call(self, func_name: str, func_address: int, call_address: int = None):
        """
        Executes the call to the given function.
        If a call hook has been provided, the hook will get run. Otherwise, instructions in the function
        will get emulated if set call_depth allows.

        :param func_name: Name of the function (or empty string)
        :param func_address: Address of function to call.
        :param call_address: Address of call instruction (if available)
        """
        # Tell context that we are currently emulating a function hook.
        # This information is important for things like pulling out function arguments out correctly.
        self.hooking_call = func_address

        try:
            # Report on function call and their arguments.
            arg_objs = self.get_function_arg_objects(func_address)
            args = [arg_obj.value for arg_obj in arg_objs]
            self.call_history.append((call_address, func_name, args))

            # Emulate the effects of any known builtin functions.
            call_hook = self.emulator.get_call_hook(func_address)
            if not call_hook:
                call_hook = self.emulator.get_call_hook(func_name)
                if not call_hook:
                    # Try one more time with a sanitized name.
                    func_name = utils.sanitize_func_name(func_name)
                    call_hook = self.emulator.get_call_hook(func_name)

            # Execute call hook if provided.
            if call_hook:
                try:
                    logger.debug(
                        "Emulating %s(%s)",
                        func_name,
                        ", ".join(f"{arg_obj.name}={hex(arg_obj.value)}" for arg_obj in arg_objs)
                    )
                    logger.debug("Running hook: %r", call_hook)
                    ret = call_hook(self, func_name, args)
                    if ret is True:
                        ret = 1
                    elif ret is False:
                        ret = 0
                    # Set return value to rax
                    if ret is not None:
                        if not isinstance(ret, int):
                            raise TypeError(f"Invalid return type. Expected 'int' but got '{type(ret)}'")
                        self.ret = ret
                except RuntimeError:
                    raise  # Allow RuntimeError exceptions to be thrown.
                except Exception as e:
                    logger.debug("Failed to emulate builtin function: %s() with error: %s", func_name, e)

            # Otherwise, emulate the function instructions if set call_depth allows.
            elif self._call_depth:
                try:
                    func = Function(func_address)
                except AttributeError as e:
                    logger.warning("Failed to execute call at 0x%X : %s", func_address, e)
                    return
                call_depth = self._call_depth
                sp_start = self._sp_start
                try:
                    self.execute(start=func.start_ea, end=func.end_ea, call_depth=call_depth - 1)
                finally:
                    # Reset after we leave call.
                    self._call_depth = call_depth
                    self._sp_start = sp_start

        finally:
            self.hooking_call = None

    @property
    def func_calls(self) -> dict:
        warnings.warn(f".func_calls is deprecated. Please use .call_history instead.", DeprecationWarning)
        # Original .func_calls overwrote calls at the same instruction. So pulling the last one to replicate this logic.
        return {address: (func_name, args) for address, func_name, args in self.call_history}

    def get_call_history(self, func_name_or_ea) -> List[Tuple[int, List]]:
        """
        Returns the call history for a specific function name.

        :returns: List of tuples containing: (ea of call, list of function arguments)
        """
        if isinstance(func_name_or_ea, str):
            func_name = func_name_or_ea
        else:
            ea = func_name_or_ea
            func_name = utils.get_function_name(ea)
        return [(address, args) for address, _func_name, args in self.call_history if _func_name == func_name]

    def prep_for_branch(self, bb_start_ea):
        """
        Modify this current context in preparation for a specific path.
        """
        if self.jcccontext.is_alt_branch(bb_start_ea):
            logger.debug("Modifying context for branch at 0x%08X", bb_start_ea)
            # Set the destination operand relative to the current context
            # to a valid value that makes this branch true.
            dst_opnd = self.jcccontext.alt_branch_data_dst
            dst_opnd = self.get_operands(ip=dst_opnd.ip)[dst_opnd.idx]
            dst_opnd.value = self.jcccontext.alt_branch_data

        self.jcccontext = JccContext()

    def get_instruction(self, ip=None) -> Instruction:
        """
        Gets the Instruction object for the current instruction pointed by the instruction pointer.

        :param ip: location of instruction pointer to pull Instruction from (default to current ip in context)
        :return: Instruction object
        """
        if ip is None:
            ip = self.ip
        return self._instruction_class(self, ip)

    @property
    def instruction(self) -> Instruction:
        return self.get_instruction()

    def get_operands(self, ip=None) -> List[Operand]:
        """
        Gets the Operand objects of all operands in the current instruction and returns them in a list.

        :param int ip: location of instruction pointer to pull operands from (defaults to current rip in context)

        :return: list of Operand objects
        """
        return self.get_instruction(ip=ip).operands

    @property
    def operands(self) -> List[Operand]:
        return self.get_operands()

    def reg_read(self, reg):
        """
        Read a register value

        >>> cpu_context = ProcessorContext()
        >>> cpu_context.reg_read("EIP")

        :param str reg: register name to be read

        :return int: value contained in specified register as int
        """
        return self.registers[reg.lower()]

    def reg_write(self, reg, val):
        """
        Write a register value

        :param str reg: register name to be written

        :param int val: value to be written to register as an int of width of the register (will be truncated as necessary)
        """
        self.registers[reg.lower()] = val

    def mem_alloc(self, size):
        """
        Allocates heap region with size number of bytes.

        :param size: Number of bytes to allocate.
        :return: starting address of allocated memory.
        """
        return self.memory.alloc(size)

    def mem_realloc(self, address, size):
        """
        Reallocates heap region with size number of bytes.

        :param address: base address to reallocate.
        :param size: Number of bytes to allocate.
        :return: address of the reallocated memory block.
        """
        new_address = self.memory.realloc(address, size)
        # Record a memory copy if pointer has changed.
        if new_address != address:
            self.memory_copies[self.ip].append((address, new_address, size))
        return new_address

    def mem_copy(self, src, dst, size):
        """
        Copy data from src address to dst address
        (Use this over mem_read/mem_write in order to allow the context to keep track of memory pointer history.)

        :param src: Source address
        :param dst: Destination address
        :param size: Number of bytes to copy over.
        :return:
        """
        self.memory_copies[self.ip].append((src, dst, size))
        self.mem_write(dst, self.mem_read(src, size))

    def get_pointer_history(self, ea):
        """
        Retrieves the history of a specific pointer.
        :param ea: Pointer to start with.
        :return: list of tuples containing (address of the memory copy, source pointer)
            - sorted by earliest to latest incarnation of the pointer. (not including itself)
        """
        history = []
        for ip, copies in sorted(list(self.memory_copies.items()), reverse=True):
            for src, dst, size in sorted(copies, reverse=True):
                if dst == ea:
                    history.append((ip, src))
                    ea = src
        history.reverse()
        return history

    def get_original_location(self, addr):
        """
        Retrieves the original location for a given address by looking through it's pointer history.

        :param addr: address of interest

        :return: a tuple containing:
            - instruction pointer where the original location was first copied
                or None if given address is already loaded or the original location could not be found.
            - either a loaded address, a tuple containing (frame_id, stack_offset) for a stack variable,
                or None if the original location could not be found.
        """
        # TODO: Consider refactoring.

        # Pull either the first seen loaded address or last seen stack variable.
        if idc.is_loaded(addr):
            return None, addr
        ip = None

        var = self.variables.get(addr, None)
        for ip, ea in reversed(self.get_pointer_history(addr)):
            if idc.is_loaded(ea):
                return ip, ea
            var = self.variables.get(ea, var)

        if var and var.is_stack:
            return ip, (var.frame_id, var.stack_offset)
        else:
            return ip, None

    def mem_read(self, address, size):
        """
        Read memory at the specified address of size size

        :param int address: address to read memory from
        :param int size: size of data to be read
        :return bytes: read data as bytes
        """
        return self.memory.read(address, size)

    def mem_write(self, address, data):
        """
        Write content contained in data to specified address

        :param int address: address to write data at
        :param bytes data: data to be written as bytes
        """
        self.memory.write(address, data)

    def mem_find(self, value, start=0, end=None):
        return self.memory.find(value, start=start, end=end)

    def mem_find_in_segment(self, value, seg_name_or_ea):
        return self.memory.find_in_segment(value, seg_name_or_ea)

    def mem_find_in_heap(self, value):
        return self.memory.find_in_heap(value)

    # TODO: Move this into Memory class and automatically decode values.
    def read_data(self, addr, size=None, data_type=None):
        """
        Reads memory at the specified address, of the specified size and convert
        the resulting data into the specified type.

        :param int addr: address to read data from
        :param int size: size of data to read
        :param data_type: type of data to be extracted
            (default to BYTE_STRING is size provided or STRING if not)
        """
        if not data_type:
            data_type = STRING if size is None else BYTE_STRING
        if size is None:
            size = 0

        if data_type == STRING:
            null_offset = self.memory.find(b"\0", start=addr)
            # It should always eventually find a null since unmapped pages
            # are all null. If we get -1 we have a bug.
            assert null_offset != -1, "Unable to find a null character!"
            return self.memory.read(addr, null_offset - addr)

        elif data_type == WIDE_STRING:
            # Step by 2 bytes to find 2 nulls on an even alignment.
            # (This helps prevent the need to take endianness into account.)
            null_offset = addr
            while self.memory.read(null_offset, 2) != b"\0\0":
                null_offset += 2

            return self.memory.read(addr, null_offset - addr)

        elif data_type == BYTE_STRING:
            return self.memory.read(addr, size)

        elif data_type == BYTE:
            return utils.struct_unpack(self.mem_read(addr, 1))

        elif data_type == WORD:
            return utils.struct_unpack(self.mem_read(addr, 2))

        elif data_type == DWORD:
            return utils.struct_unpack(self.mem_read(addr, 4))

        elif data_type == QWORD:
            return utils.struct_unpack(self.mem_read(addr, 8))

        raise ValueError("Invalid data_type: {!r}".format(data_type))

    def write_data(self, addr, value, data_type=None):
        """
        Writes memory at the specified address after converting the value
        into data based on the specified data type.

        :param int addr: address to write data to
        :param value: integer or byte string to write
        :param data_type: type of data to convert value from.
            (defaults to BYTE_STRING, STRING, or DWORD based on input data)
        """
        if not data_type:
            if isinstance(value, str):
                data_type = STRING
            elif isinstance(value, bytes):
                data_type = BYTE_STRING
            elif isinstance(value, int):
                data_type = DWORD
            else:
                raise ValueError(f"Invalid data type: {type(value)}")

        if data_type == BYTE_STRING:
            data = value

        elif data_type == STRING:
            data = value
            if isinstance(data, str):
                data = data.encode("utf8")
            data += b"\0"

        elif data_type == WIDE_STRING:
            data = value
            if isinstance(data, str):
                data = data.encode("utf-16-le")
            data += b"\0\0"

        elif data_type == BYTE:
            data = bytes([value])

        elif data_type == WORD:
            data = utils.struct_pack(value, width=2)

        elif data_type == DWORD:
            data = utils.struct_pack(value, width=4)

        elif data_type == QWORD:
            data = utils.struct_pack(value, width=8)

        else:
            raise ValueError(f"Invalid data_type: {repr(data_type)}")

        self.mem_write(addr, data)

    def get_function_signature(self, func_ea=None, force=False, num_args=None) -> Optional[FunctionSignature]:
        """
        Returns the function signature of the given func_ea with argument values pulled
        from this context.

        :param int func_ea: address of the function to pull signature from.
            The first operand is used if not provided. (helpful for a "call" instruction)
        :param bool force: Whether to force a function signature using cdecl calling
            convention and no arguments if we fail to generate the signature.
            (Useful when trying to declare a function that was dynamically created in a register)
        :param int num_args: Force a specific number of arguments in the signature.
            If not provided, number of arguments is determined by the disassembler.
            Extra arguments not defined by the disassembler are assumed to be 'int' type.
            Avoid using num_args and adjust the returned FunctionSignature manually
            if more customization is needed.
            (NOTE: The function signature will be forced on failure if this is set.)

        :return: FunctionSignature object or None if not applicable

        :raises RuntimeError: If a function signature could not be created from given ea.
        :raises ValueError: If num_args is negative
        """
        # If func_ea is not given, assume we are using the first operand from a call instruction.
        if not func_ea:
            if not self.operands:
                return None
            operand = self.operands[0]
            # function pointer can be a memory reference or immediate.
            func_ea = operand.addr or operand.value
        else:
            operand = None

        force = force or num_args is not None

        try:
            func_sig = FunctionSignature(self, func_ea, operand=operand)
        except RuntimeError as e:
            # If we fail to get a function signature but force is set, set the type to
            # cdecl with no arguments.
            if force:
                logger.warning(
                    "Failed to create function signature at 0x{:0X} with error: {}\n"
                    "Forcing signature with assumed cdecl calling convention.".format(func_ea, e)
                )
                idc.SetType(func_ea, "int __cdecl no_name();")
                func_sig = FunctionSignature(self, func_ea)
            else:
                raise

        if num_args is not None:
            if num_args < 0:
                raise ValueError("num_args is negative")
            arg_types = func_sig.arg_types
            if len(arg_types) > num_args:
                func_sig.arg_types = arg_types[:num_args]
            elif len(arg_types) < num_args:
                func_sig.arg_types = arg_types + ("int",) * (num_args - len(arg_types))

        return func_sig

    def get_function_arg_objects(self, func_ea=None, num_args=None) -> List[FunctionArg]:
        """
        Returns the FunctionArg objects for this context based on the
        given function.

        >>> cpu_context = ProcessorContext()
        >>> args = cpu_context.get_function_arg_objects(0x180011772)

        :param int func_ea: Ea of the function to pull a signature from.
        :param int num_args: Force a specific number of arguments.
            If not provided, number of arguments is determined by the disassembler.
            Extra arguments not defined by the disassembler are assumed to be 'int' type.
            Use get_function_signature() and adjust the FunctionSignature manually
            if more customization is needed.
            (NOTE: The function signature will be forced on failure if this is set.)

        :returns: list of FunctionArg objects
        """
        func_sig = self.get_function_signature(func_ea, num_args=num_args)
        if not func_sig:
            return []

        return func_sig.args

    # TODO: Replace this function with get_function_arg_objects()
    def get_function_args(self, func_ea=None, num_args=None) -> List[int]:
        """
        Returns the FunctionArg values for this context based on the given function.
        """
        args = self.get_function_arg_objects(func_ea=func_ea, num_args=num_args)
        return [arg.value for arg in args]

    @property
    def function_args(self) -> List[FunctionArg]:
        """
        The function arguments currently set based on the function in the first operand.
        """
        return self.get_function_arg_objects()

    @property
    def passed_in_args(self) -> List[FunctionArg]:
        """
        The function arguments for the current function.
        """
        func = ida_funcs.get_func(self.ip)
        return self.get_function_arg_objects(func.start_ea)

    @property
    def files(self) -> List[File]:
        """
        All File objects in the current context.
        """
        return list(self.objects.query(File))

    def get_file(self, handle_or_path: Union[int, str], default=None) -> File:
        """
        Gets a file by handle or path. Returns default if not existent.

        :raises TypeError: if handle points and object that is not a File.
        """
        warnings.warn(
            "get_file() is deprecated. Please use objects.query() instead.",
            DeprecationWarning
        )
        if isinstance(handle_or_path, int):
            condition = dict(handle=handle_or_path)
        else:
            condition = dict(path=handle_or_path)

        for obj in self.objects.query(File, **condition):
            return obj
        return default

    @property
    def regkeys(self) -> List[RegKey]:
        """
        The opened registry keys for this context.
        """
        return list(self.objects.query(RegKey))

    @property
    def services(self) -> List[Service]:
        """
        The created services for this context.
        """
        return list(self.objects.query(Service))



"""
Interface for creating register families.
"""

from builtins import object
from copy import deepcopy


class Register(object):
    """
    Provides access to a register family.

    :param size int: size of register in bytes
    :param **masks: maps member names to a mask of the register value it corresponds to.

    >>> reg = Register(8, rax=0xFFFFFFFFFFFFFFFF, eax=0xFFFFFFFF, ax=0xFFFF, al=0xFF, ah=0xFF00)
    >>> reg.rax
    0
    >>> reg.ax
    0
    >>> reg.ah = 0x23
    >>> reg.ah
    0x23
    >>> reg.ax
    0x2300
    >>> reg.eax
    0x00002300
    >>> reg.eax = 0x123
    >>> reg.al
    0x23
    >>> reg.ah
    0x01
    >>> reg.rax
    0x0000000000000123
    """

    def __init__(self, size, **masks):
        # We are modifying self.__dict__ directly to avoid triggering the
        # overwritten __setattr__()
        self_dict = self.__dict__
        self_dict["size"] = size
        self_dict["_size_mask"] = 2 ** (8 * size) - 1
        self_dict["_value"] = 0

        _masks = {}
        for name, mask in list(masks.items()):
            # Get position of rightmost set bit in mask
            shift = 0
            if mask:
                _mask = mask
                while not _mask & 0x1:
                    _mask >>= 1
                    shift += 1
            _masks[name.lower()] = (mask, shift)
        self_dict["_masks"] = _masks

    def __deepcopy__(self, memo):
        copy = self.__new__(self.__class__)
        memo[id(self)] = copy
        copy_dict = copy.__dict__
        copy_dict["size"] = self.size
        copy_dict["_size_mask"] = self._size_mask
        copy_dict["_value"] = self._value
        copy_dict["_masks"] = dict(self._masks)
        return copy

    def __getattr__(self, reg_name):
        try:
            mask, shift = self._masks[reg_name]
        except KeyError:
            raise AttributeError("Invalid register name: {}".format(reg_name))
        return (self._value & mask) >> shift

    def __getitem__(self, reg_name):
        return self.__getattr__(reg_name)

    def __setattr__(self, reg_name, value):
        try:
            mask, shift = self._masks[reg_name]
        except KeyError:
            raise AttributeError("Invalid register name: {}".format(reg_name))
        if not isinstance(value, int):
            raise ValueError("Register value must be int or long, got {}".format(type(value)))
        self.__dict__["_value"] = (self._value & (mask ^ self._size_mask)) | ((value & (mask >> shift)) << shift)

    def __setitem__(self, reg_name, value):
        self.__setattr__(reg_name, value)

    def __contains__(self, reg_name):
        return reg_name in self._masks

    @property
    def names(self):
        return list(self._masks.keys())


class RegisterMap(object):
    """
    Holds register families and allows for direct access.

    This class contains all the CPU registers.  It is updated by both the CPU class, which
    updates the main CPU registers and the Processor class, which will update FLAGS.
    """

    def __init__(self, registers):
        """
        :param registers: list of Register instances
        """
        self_dict = self.__dict__
        self_dict["_registers"] = registers
        self_dict["_reg_map"] = self._build_reg_map(registers)

    @staticmethod
    def _build_reg_map(registers):
        """Builds and returns a dictionary mapping register names to their respective Register object."""
        # Build a hash table mapping member names to registers.
        # (This also validates that we have no collisions while we are at it.)
        reg_map = {}
        for register in registers:
            for name in register.names:
                if name in reg_map:
                    raise RuntimeError("Duplicate register name: {}".format(name))
                reg_map[name] = register
        return reg_map

    def __deepcopy__(self, memo):
        copy = self.__new__(self.__class__)
        memo[id(self)] = copy

        copy_dict = copy.__dict__
        copy_dict["_registers"] = [deepcopy(reg, memo) for reg in self._registers]
        copy_dict["_reg_map"] = self._build_reg_map(copy._registers)

        return copy

    def __getattr__(self, reg_name):
        try:
            register = self._reg_map[reg_name]
        except KeyError:
            raise AttributeError("Invalid register: {}".format(reg_name))
        return register[reg_name]

    def __getitem__(self, reg_name):
        return self.__getattr__(reg_name.lower())

    def __setattr__(self, reg_name, value):
        try:
            register = self._reg_map[reg_name]
        except KeyError:
            raise AttributeError("Invalid register: {}".format(reg_name))
        register[reg_name] = value

    def __setitem__(self, reg_name, value):
        self.__setattr__(reg_name.lower(), value)

    @property
    def names(self):
        return list(self._reg_map.keys())



"""
CPU Emulator Built-in Windows API functions

These functions are used to emulate the effects of "useful" Windows API functions.

Add any builtin functions that need to be handled below.  The function should be declared as such

# Using the same function for multiple API functions
@builtin_func("GetSystemDirectoryA")
@builtin_func("GetSystemDirectoryW")
#typespec("UINT GetSystemDirectoryA(char* lpBuffer, UINT uSize);")
def memmove(cpu_context, func_name, func_args):
    logger.debug("Emulating GetSystemDirectoryA")

"""

import logging
import random

from . import win_constants as wc
from ... import constants
from ...call_hooks import builtin_func
from ... import actions

logger = logging.getLogger(__name__)


@builtin_func("ShellExecuteA")
@builtin_func("ShellExecuteW")
#typespec("HINSTANCE ShellExecuteA(HWND hwnd, LPCSTR lpOperation, LPCSTR lpFile, LPCSTR lpParameters, LPCSTR lpDirectory, INT nShowCmd);)")
def shell_execute(cpu_context, func_name, func_args):
    """
    Performs an operation on a specified file
    """
    wide = func_name.endswith("W")

    _, operation_ptr, file_ptr, params_ptr, dir_ptr, visibility = func_args

    filepath = cpu_context.read_data(
        file_ptr, data_type=constants.WIDE_STRING if wide else constants.STRING
    ).decode("utf-16-le" if wide else "utf8")

    if params_ptr:
        params = cpu_context.read_data(
            params_ptr, data_type=constants.WIDE_STRING if wide else constants.STRING
        ).decode("utf-16-le" if wide else "utf8")
    else:
        params = u""

    if operation_ptr:
        operation = cpu_context.read_data(
            operation_ptr, data_type=constants.WIDE_STRING if wide else constants.STRING
        ).decode("utf-16-le" if wide else "utf8")
    else:
        operation = u""

    if dir_ptr:
        directory = cpu_context.read_data(
            dir_ptr, data_type=constants.WIDE_STRING if wide else constants.STRING
        ).decode("utf-16-le" if wide else "utf8")
    else:
        directory = u""

    logger.debug("%s: %r %r", func_name, operation, filepath)
    cpu_context.actions.add(
        actions.ShellOperation(cpu_context.ip, operation, filepath, params, directory, wc.Visibility(visibility))
    )

    return random.randint(wc.MIN_HANDLE, wc.MAX_HANDLE)


# Dictionary mapping CSIDL values to their default values on 64-bit Windows 10. Created by taking the listing from the
# SpecialFoldersView.exe utility from NirSoft and then cleaning it up to use more appropriate environment variables.
CSIDL_DICT_WIN10x64 = {
    wc.CSIDL.CSIDL_PROGRAMS: r"%AppData%\Microsoft\Windows\Start Menu\Programs",
    wc.CSIDL.CSIDL_PERSONAL: r"%UserProfile%\Documents",
    wc.CSIDL.CSIDL_FAVORITES: r"%UserProfile%\Favorites",
    wc.CSIDL.CSIDL_STARTUP: r"%AppData%\Microsoft\Windows\Start Menu\Programs\Startup",
    wc.CSIDL.CSIDL_RECENT: r"%AppData%\Microsoft\Windows\Recent",
    wc.CSIDL.CSIDL_SENDTO: r"%AppData%\Microsoft\Windows\SendTo",
    wc.CSIDL.CSIDL_STARTMENU: r"%AppData%\Microsoft\Windows\Start Menu",
    wc.CSIDL.CSIDL_MYMUSIC: r"%UserProfile%\Music",
    wc.CSIDL.CSIDL_MYVIDEO: r"%UserProfile%\Videos",
    wc.CSIDL.CSIDL_DESKTOPDIRECTORY: r"%UserProfile%\Desktop",
    wc.CSIDL.CSIDL_NETHOOD: r"%AppData%\Microsoft\Windows\Network Shortcuts",
    wc.CSIDL.CSIDL_FONTS: r"%SystemRoot%\Fonts",
    wc.CSIDL.CSIDL_TEMPLATES: r"%AppData%\Microsoft\Windows\Templates",
    wc.CSIDL.CSIDL_COMMON_STARTMENU: r"%AllUsersProfile%\Microsoft\Windows\Start Menu",
    wc.CSIDL.CSIDL_COMMON_PROGRAMS: r"%AllUsersProfile%\Microsoft\Windows\Start Menu\Programs",
    wc.CSIDL.CSIDL_COMMON_STARTUP: r"%AllUsersProfile%\Microsoft\Windows\Start Menu\Programs\Startup",
    wc.CSIDL.CSIDL_COMMON_DESKTOPDIRECTORY: r"%Public%\Desktop",
    wc.CSIDL.CSIDL_APPDATA: r"%AppData%",
    wc.CSIDL.CSIDL_PRINTHOOD: r"%AppData%\Microsoft\Windows\Printer Shortcuts",
    wc.CSIDL.CSIDL_LOCAL_APPDATA: r"%LocalAppData%",
    wc.CSIDL.CSIDL_COMMON_FAVORITES: r"%UserProfile%\Favorites",
    wc.CSIDL.CSIDL_INTERNET_CACHE: r"%LocalAppData%\Microsoft\Windows\INetCache",
    wc.CSIDL.CSIDL_COOKIES: r"%LocalAppData%\Microsoft\Windows\INetCookies",
    wc.CSIDL.CSIDL_HISTORY: r"%LocalAppData%\Microsoft\Windows\History",
    wc.CSIDL.CSIDL_COMMON_APPDATA: r"%AllUsersProfile%",
    wc.CSIDL.CSIDL_WINDOWS: r"%WinDir%",
    wc.CSIDL.CSIDL_SYSTEM: r"%SystemRoot%\System32",
    wc.CSIDL.CSIDL_PROGRAM_FILES: r"%ProgramFiles%",
    wc.CSIDL.CSIDL_MYPICTURES: r"%UserProfile%\Pictures",
    wc.CSIDL.CSIDL_PROFILE: r"%UserProfile%",
    wc.CSIDL.CSIDL_SYSTEMX86: r"%SystemRoot%\SysWOW64",
    wc.CSIDL.CSIDL_PROGRAM_FILESX86: r"%ProgramFiles(x86)%",
    wc.CSIDL.CSIDL_PROGRAM_FILES_COMMON: r"%CommonProgramFiles%",
    wc.CSIDL.CSIDL_PROGRAM_FILES_COMMONX86: r"%CommonProgramFiles(x86)%",
    wc.CSIDL.CSIDL_COMMON_TEMPLATES: r"%AllUsersProfile%\Microsoft\Windows\Templates",
    wc.CSIDL.CSIDL_COMMON_DOCUMENTS: r"%Public%\Documents",
    wc.CSIDL.CSIDL_COMMON_ADMINTOOLS: r"%AllUsersProfile%\Microsoft\Windows\Start Menu\Programs\Administrative Tools",
    wc.CSIDL.CSIDL_ADMINTOOLS: r"%AppData%\Microsoft\Windows\Start Menu\Programs\Administrative Tools",
    wc.CSIDL.CSIDL_COMMON_MUSIC: r"%Public%\Music",
    wc.CSIDL.CSIDL_COMMON_PICTURES: r"%Public%\Pictures",
    wc.CSIDL.CSIDL_COMMON_VIDEO: r"%Public%\Videos",
    wc.CSIDL.CSIDL_CDBURN_AREA: r"%LocalAppData%\Microsoft\Windows\Burn\Burn"
}


@builtin_func("SHGetFolderPathA")
@builtin_func("SHGetFolderPathW")
@builtin_func("SHGetSpecialFolderPathA")
@builtin_func("SHGetSpecialFolderPathW")
#typespec("SHFOLDERAPI SHGetFolderPathA(HWND hwnd, int csidl, HANDLE hToken, DWORD dwFlags,  LPSTR  pszPath);")
#typespec("BOOL SHGetSpecialFolderPathA(HWND hwnd, LPSTR pszPath, int csidl, BOOL fCreate);")
def sh_get_folder_path(cpu_context, func_name, func_args):
    """
    Retrieves the CSIDL folder mapping for the provided CSIDL Enumeration

    Attempts to expand folder based on default 64-bit Windows 10 values.
    """
    wide = func_name.endswith("W")
    if "Special" in func_name:
        _, path_ptr, csidl, _ = func_args
    else:
        _, csidl, _, _, path_ptr = func_args

    # Mask off any CSIDL flags
    csidl &= 0xff
    if csidl in CSIDL_DICT_WIN10x64:
        folder = CSIDL_DICT_WIN10x64[csidl]
    else:
        logger.warning("Could not acquire a folder for CSIDL %d", csidl)
        folder = f"{csidl}"

    logger.debug("Writing CSIDL %s to 0x%08X", folder, path_ptr)
    cpu_context.write_data(path_ptr, folder, data_type=constants.WIDE_STRING if wide else constants.STRING)
    return wc.ERROR_SUCCESS



"""
Interface for registering and accessing decoders.
"""

from collections import namedtuple
import os
import logging

import pkg_resources

from .decoder import Decoder


logger = logging.getLogger(__name__)
Source = namedtuple("Source", ("name", "path"))
# Set of decoder source names mapped to a directory path.
_sources = {}
_default_source = None


def get_default_source():
    return _default_source


def get_sources():
    """
    Returns list of Source namedtuples containing name and path.
    """
    return _sources.values()


def set_default_source(source_name):
    """
    Sets a default parser source to use if not explicitly defined.
    If this is not set, all sources will be considered.

    :param source_name: The name of the source to set.
    """
    global _default_source
    _default_source = source_name


def clear_default_source():
    """
    Clears a previously set default source.
    """
    global _default_source
    _default_source = None


def register_entry_points():
    """
    Registers decoders found in entry_point: "kordesii.decoders"
    """
    for entry in pkg_resources.iter_entry_points("kordesii.decoders"):
        package = entry.load()
        register_decoder_package(package, source_name=entry.name)


def register_decoder_directory(directory, source_name=None):
    """
    Registers decoders found in directory. This function allows you to register one-off decoders
    that are not part of an installed python package.

    :param str directory: An extra directory to look for one-off decoders.
    :param source_name: Unique name to give to the source. (uses directory path otherwise)

    :raises ValueError: If loaded config file is invalid.
    """
    global _sources

    if not os.path.isdir(directory):
        raise ValueError(u"Decoder directory not found or not a directory: {!r}".format(directory))

    # # Ensure this directory can be converted to a package and pull config_file_path if available.
    # package = _create_package(directory)

    if not source_name:
        source_name = directory

    _sources[source_name] = Source(source_name, directory)


def register_decoder_package(package, source_name=None):
    """
    Registers Python package containing Kordesii decoders.

    :param package: An Python package containing submodules that contain Kordesii decoders.
        NOTE: Package must be discoverable in subprocesses without modifying the python path.
              Please use register_decoder_directory() instead if that is not possible.
    :param source_name: Unique name to give to the source. (uses package name otherwise)
    """
    if not hasattr(package, "__path__"):
        raise ValueError(u"{!r} is not a Python package".format(package))

    if not source_name:
        source_name = package.__name__.lower()

    # We must register as directories because we will eventually to need
    # to get the script path to run it.
    register_decoder_directory(os.path.dirname(package.__file__), source_name=source_name)


def iter_decoders(name=None, source=None):
    """
    Iterates paths to all registered decoders.

    :param str name: Filters decoder based on a particular name. (":" notation is also supported)
    :param str source: Filters decoder based on a particular source.
                       (source is either the name of a python package or path to local directory)

    :yields: tuple containing: (Source tuple, decoder_path)

    :raises ValueError: If a decoder name or source could not be found.
    """
    global _sources

    if name and not source:
        # If name is using ":" notation, assume it is being organized by "source_name:decoder_name"
        # (os.path.basename is necessary in-case source is a file path containing ":"'s)
        orig_name = name
        _, _, name = os.path.basename(name).rpartition(":")
        source = orig_name[: -(len(name) + 1)]

    # Use default source if one is not provided.
    source = source or _default_source or None

    sources = []
    if source:
        if source in _sources:
            sources.append((source, _sources[source]))
    else:
        sources += _sources.items()

    for source_name, source in sources:
        # Get script path for decoder.
        if name:
            # Pull script using a "." notation to indicate subpackages.
            script_path = os.path.join(source.path, *name.split(".")) + ".py"
            # Also try with legacy postfix if it doesn't exists.
            if not os.path.exists(script_path):
                script_path = script_path[:-3] + "_StringDecode.py"
            if os.path.exists(script_path):
                yield Decoder(script_path, name=name, source=source)
            else:
                logger.debug("Unable to find {}:{} decoder.".format(source_name, name))
        else:
            # Extract all decoders within the directory
            for root, directories, filenames in os.walk(source.path):
                for filename in sorted(filenames, key=lambda f: f.lower()):  # Case-insensitive sorting.
                    if filename.endswith(".py") and not filename.startswith("_"):
                        script_path = os.path.join(root, filename)
                        rel_path, _ = os.path.splitext(os.path.relpath(script_path, source.path))
                        _name = rel_path.replace(os.path.sep, ".")
                        yield Decoder(script_path, name=_name, source=source)


def get_decoder_descriptions(name=None, source=None):
    """
    Retrieve list of decoder descriptions

    :param str name: Filters parser based on a particular name. (":" notation is also supported)
    :param str source: Filters parser based on a particular source.
                       (source is either the name of a python package or path to local directory)

    Returns list of tuples per parser. Tuple contains parser name, author, and description.
    """
    descriptions = []
    for decoder in iter_decoders(name=name, source=source):
        descriptions.append((decoder.name, decoder.source.name, decoder.author, decoder.description))
    return sorted(descriptions, key=lambda e: tuple(sub.lower() for sub in e))  # Case-insensitive sorting.



"""
Import built-in function code based on input file.

These functions are used to emulate the effects of known builtin functions.

Add any builtin functions that need to be handled below.  The function should be declared as such

# Using the same function for multiple instructions:
@builtin_func("memmove")
@builtin_func("memcpy")
def _memcpy(cpu_context, call_ip, func_name, func_args):
    print "IN memmove or memcpy"
    return 1  # Return anything to be placed into rax (or equivalent)

# Using a single function for a builtin
@builtin_func
def memmove(cpu_context, call_ip, func_name, func_args):
    print "IN memmove"

"""

from ..registry import registrar

# Dictionary containing builtin function names -> function
BUILTINS = {}

# Dictionary containing builtin function names -> function
builtin_func = registrar(BUILTINS, name="builtin")


from . import stdlib

# TODO: Call the framework file type information rather than the IDA functions.
import idc

file_type = idc.get_inf_attr(idc.INF_FILETYPE)
if file_type == idc.FT_PE:
    from . import win_api
elif file_type == idc.FT_ELF:
    pass



"""
Interface for operand management in ARM.
"""

import logging
from typing import Optional, List

import ida_ua

from kordesii.utils.function_tracing import utils
from kordesii.utils.function_tracing.ARM import ida_arm, utils as arm_utils
from kordesii.utils.function_tracing.exceptions import FunctionTracingError
from kordesii.utils.function_tracing.operands import Operand

logger = logging.getLogger(__name__)


class ARMOperand(Operand):

    _shift_map = {
        ida_arm.shift_t.LSL: arm_utils.lsl,
        ida_arm.shift_t.LSR: arm_utils.lsr,
        ida_arm.shift_t.ASR: arm_utils.asr,
        ida_arm.shift_t.ROR: arm_utils.ror,
        # rrx requires also passing in carry. (see usage)
        # TODO: Support other shift operations reported in IDA.
        ida_arm.shift_t.UXTX: arm_utils.lsl
    }

    @property
    def has_register(self):
        return super().has_register or self.type in (
            ida_arm.o_shreg, ida_arm.o_reglist, ida_arm.o_creg, ida_arm.o_creglist, ida_arm.o_fpreglist
        )

    @property
    def base(self) -> Optional[int]:
        """The value of the base register if operand is a displacement."""
        if not self.has_phrase:
            return None
        value = self._cpu_context.registers[utils.reg2str(self._op.reg)]
        return utils.signed(value)

    @base.setter
    def base(self, value: int):
        """Sets the value of the base register if operand is a displacement."""
        if not self.has_phrase:
            return
        self._cpu_context.registers[utils.reg2str(self._op.reg)] = value

    @property
    def offset(self) -> Optional[int]:
        """The offset value if the operand is a displacement/phrase."""
        # [R1, R2]
        if self.type == ida_ua.o_phrase:
            second_reg = ida_arm.secreg(self._op)  # pulling the R2
            offset = self._cpu_context.registers[utils.reg2str(second_reg)]
            # We could also have a shift applied in the offset.
            #   [R1, R2, LSL #3]
            offset = self._calc_shift(offset)
            return utils.signed(offset)

        # [R1, #1]
        elif self.type == ida_ua.o_displ:
            return utils.signed(self._op.addr)

        return None

    @property
    def shift_count(self) -> int:
        """The amount to shift (if a shifted register)"""
        if self.type == ida_ua.o_phrase:
            # For a phrase, shift is in op.value
            return self._op.value

        if self.type == ida_arm.o_shreg:
            # Shift can be an immediate or another register.
            # I believe we determine which one it is based on whether op.value is zero.
            count = self._op.value
            if not count:
                count = self._cpu_context.registers[utils.reg2str(ida_arm.secreg(self._op))]
            return count

        return 0  # not shifted

    def _calc_shift(self, value) -> int:
        """
        Calculates the shift applied within the operand.
        This could be applied directly or within the offset of a displacement.
            e.g.
                R2, LSL #3 -> R2 << 3
                R2, LSL R3 -> R2 << R3
                [R1, R2, LSL #3]  -> R2 << 3

        NOTE: Any modifications to the carry flag will be applied only if
             the condition flag is set and the context's ip is the same as the
             address of the operand's instruction.

        :param value: The base value the shift is to be applied to.
        :return: Results of the shift (or original value back if no shift is applicable.)
        """
        count = self.shift_count
        if count > 0:
            shift_op = self._op.specflag2
            if shift_op == ida_arm.shift_t.RRX:  # RRX also requires original carry flag
                carry, value = arm_utils.rrx(self._cpu_context.registers.c, value, count)
            else:
                carry, value = self._shift_map[shift_op](value, count)

            # Update carry flag if condition flag is set for instruction (S postfix)
            # (But only update if context's instruction pointer is still looking at this instruction.)
            if self._insn.auxpref & ida_arm.aux_cond and self._cpu_context.ip == self.ip:
                self._cpu_context.registers.c = carry

        return value

    @property
    def addr(self) -> Optional[int]:
        """
        The referenced memory address of the operand.
        :return int: Memory address or None if operand is not a memory reference.
        """
        addr = None
        if self.has_phrase:
            addr = self.base

            # Ignore including the offset if post indexed.
            #   ie. include offset for [R2, #4] but ignore for [R2], #4
            if not self._insn.auxpref & ida_arm.aux_postidx:
                offset = self.offset
                logger.debug("0x%X + 0x%X = 0x%X", addr, offset, addr + offset)
                addr += self.offset

            if addr < 0:
                logger.debug("Address is negative, resorting to address of 0.")
                addr = 0

            self._record_stack_variable(addr)

        elif self.type == ida_ua.o_mem:
            addr = self._op.addr
            # Record the global variable before we return.
            self._cpu_context.variables.add(addr, reference=self.ip)

        if addr is not None:
            logger.debug("&%s -> 0x%X", self.text, addr)

        return addr

    @property
    def is_signed(self) -> bool:
        """
        Whether the memory addressing is signed.
        """
        return bool(self._insn.auxpref & (ida_arm.aux_sb | ida_arm.aux_sh | ida_arm.aux_sw))

    @property
    def is_register_list(self) -> bool:
        """
        Whether operand is a register list.
        """
        return self.type == ida_arm.o_reglist

    @property
    def register_list(self) -> Optional[List[str]]:
        """
        List of register names if operand is a register list.
        """
        if not self.is_register_list:
            return None
        # Register numbers are stored in a bitmap in specval.
        reg_bitmap = self._op.specval
        return [utils.reg2str(i) for i in range(16) if reg_bitmap & (1 << i)]

    @property
    def value(self):
        # Barrel shifter
        if self.type == ida_arm.o_shreg:
            value = self._cpu_context.registers[utils.reg2str(self._op.reg)]
            value = self._calc_shift(value)
            return value

        # Register list
        if self.type == ida_arm.o_reglist:
            return [self._cpu_context.registers[reg] for reg in self.register_list]

        value = super().value

        # If a memory reference, the final value may be signed.
        if self.is_memory_reference and self.is_signed:
            value = utils.signed(value)

        return value

    @value.setter
    def value(self, value):
        try:
            logger.debug("0x%X -> %s", value, self.text)
        except TypeError:
            logger.debug("%r -> %s", value, self.text)

        # Barrel shifter
        if self.type == ida_arm.o_shreg:
            raise FunctionTracingError(f"Unable to set value to operand with a shift", ip=self.ip)

        # Register list
        if self.type == ida_arm.o_reglist:
            reg_list = self.register_list
            # To set the value for an operand that is a register list, user must provide
            # a list of value of equal size.
            # User can use None to indicate not to update that register.
            if not isinstance(value, list) or len(value) != len(reg_list):
                raise ValueError(f"Operand value for {self.text} must be a list of {len(reg_list)} values.")
            for reg_name, reg_value in zip(reg_list, value):
                if reg_value is not None:
                    self._cpu_context.registers[reg_name] = reg_value
            return

        super(ARMOperand, self.__class__).value.__set__(self, value)



"""
Interface for interesting actions.
"""

from dataclasses import dataclass
import logging
from typing import Union, List, Optional, Iterable
from .call_hooks.win_api import win_constants

logger = logging.getLogger(__name__)


@dataclass
class Action:
    ip: int


class ActionList(object):
    """
    Represents a reverse linked list of actions that have occurred up
    to a specific ProcessorContext.
    """

    def __init__(self, *actions: Action):
        self.tail: Optional[ActionNode] = None
        for action in actions:
            self.add(action)

    def __repr__(self):
        return f"ActionList({repr(self.tail) if self.tail else ''})"

    def __deepcopy__(self, memo):
        copy = ActionList()
        copy.tail = self.tail
        return copy

    def __iter__(self):
        if self.tail:
            yield from self.tail

    def __reversed__(self):
        if self.tail:
            yield from reversed(self.tail)

    def __getitem__(self, index: int):
        return list(self)[index]

    def __len__(self):
        return len(list(self))

    def __bool__(self):
        return bool(self.tail)

    def __contains__(self, item):
        return any(item == action for action in self)

    def add(self, action: Action):
        self.tail = ActionNode(action, prev=self.tail)


class ActionNode(object):
    """
    Represents a node of a reverse linked list of actions that have occurred up
    to a specific ProcessorContext.
    """

    def __init__(self, action: Action, prev: Optional["ActionNode"] = None):
        self.action = action
        self.prev = prev

    def __repr__(self):
        if self.prev:
            return f"{self.prev!r} -> {self.action}"
        else:
            return f"{self.action}"

    def __iter__(self):
        """
        Iterates actions from the least recent action that has occurred to
        the most recent action that has occurred.
        """
        if self.prev:
            yield from self.prev
        yield self.action

    def __reversed__(self):
        """
        Iterates actions from the most recent action that has occurred to
        the least recent action that has occurred.
        """
        yield self.action
        if self.prev:
            yield from reversed(self.prev)


@dataclass
class CommandExecuted(Action):
    command: str
    visibility: win_constants.Visibility = None


@dataclass
class DirectoryCreated(Action):
    path: str


@dataclass
class FileCreated(Action):
    handle: int
    path: str
    mode: str


@dataclass
class FileOpened(Action):
    handle: int
    path: str
    mode: str


@dataclass
class FileTruncated(Action):
    handle: int
    path: str
    mode: str


@dataclass
class FileDeleted(Action):
    handle: int
    path: str


@dataclass
class FileMoved(Action):
    handle: int
    old_path: str
    new_path: str


@dataclass
class FileClosed(Action):
    handle: int


@dataclass
class FileWritten(Action):
    handle: int
    data: bytes


@dataclass
class RegKeyOpened(Action):
    handle: int
    path: str
    root_key: str
    sub_key: str


@dataclass
class RegKeyDeleted(Action):
    handle: int
    path: str


@dataclass
class RegKeyValueDeleted(Action):
    handle: int
    path: str
    value_name: str


@dataclass
class RegKeyValueSet(Action):
    handle: int
    path: str
    data_type: str
    data: Union[bytes, str, List[str], int, None]


@dataclass
class ServiceCreated(Action):
    handle: int
    name: str
    access: win_constants.ServiceAccess
    service_type: win_constants.ServiceType
    start_type: win_constants.ServiceStart
    display_name: str
    binary_path: str


@dataclass
class ServiceOpened(Action):
    handle: int
    name: str


@dataclass
class ServiceDeleted(Action):
    handle: int


@dataclass
class ServiceDescriptionChanged(Action):
    handle: int
    description: str


@dataclass
class ShellOperation(Action):
    operation: str
    path: str
    parameters: str
    directory: str
    visibility: win_constants.Visibility = None



from ..registers import Register, RegisterMap
from ..utils import get_bits


def ARM_Registers():
    """Initializes registers for ARM architecture"""
    registers = [
        Register(8, x0=0xFFFFFFFFFFFFFFFF, w0=0xFFFFFFFF, r0=0xFFFFFFFF),
        Register(8, x1=0xFFFFFFFFFFFFFFFF, w1=0xFFFFFFFF, r1=0xFFFFFFFF),
        Register(8, x2=0xFFFFFFFFFFFFFFFF, w2=0xFFFFFFFF, r2=0xFFFFFFFF),
        Register(8, x3=0xFFFFFFFFFFFFFFFF, w3=0xFFFFFFFF, r3=0xFFFFFFFF),
        Register(8, x4=0xFFFFFFFFFFFFFFFF, w4=0xFFFFFFFF, r4=0xFFFFFFFF),
        Register(8, x5=0xFFFFFFFFFFFFFFFF, w5=0xFFFFFFFF, r5=0xFFFFFFFF),
        Register(8, x6=0xFFFFFFFFFFFFFFFF, w6=0xFFFFFFFF, r6=0xFFFFFFFF),
        Register(8, x7=0xFFFFFFFFFFFFFFFF, w7=0xFFFFFFFF, r7=0xFFFFFFFF),
        Register(8, x8=0xFFFFFFFFFFFFFFFF, w8=0xFFFFFFFF, r8=0xFFFFFFFF),
        Register(8, x9=0xFFFFFFFFFFFFFFFF, w9=0xFFFFFFFF, r9=0xFFFFFFFF),
        Register(8, x10=0xFFFFFFFFFFFFFFFF, w10=0xFFFFFFFF, r10=0xFFFFFFFF),
        Register(8, x11=0xFFFFFFFFFFFFFFFF, w11=0xFFFFFFFF, r11=0xFFFFFFFF),
        Register(8, x12=0xFFFFFFFFFFFFFFFF, w12=0xFFFFFFFF, r12=0xFFFFFFFF),
        Register(8, x13=0xFFFFFFFFFFFFFFFF, w13=0xFFFFFFFF, r13=0xFFFFFFFF),
        Register(8, x14=0xFFFFFFFFFFFFFFFF, w14=0xFFFFFFFF, r14=0xFFFFFFFF),
        Register(8, x15=0xFFFFFFFFFFFFFFFF, w15=0xFFFFFFFF, r15=0xFFFFFFFF),
        Register(8, x16=0xFFFFFFFFFFFFFFFF, w16=0xFFFFFFFF),
        Register(8, x17=0xFFFFFFFFFFFFFFFF, w17=0xFFFFFFFF),
        Register(8, x18=0xFFFFFFFFFFFFFFFF, w18=0xFFFFFFFF),
        Register(8, x19=0xFFFFFFFFFFFFFFFF, w19=0xFFFFFFFF),
        Register(8, x20=0xFFFFFFFFFFFFFFFF, w20=0xFFFFFFFF),
        Register(8, x21=0xFFFFFFFFFFFFFFFF, w21=0xFFFFFFFF),
        Register(8, x22=0xFFFFFFFFFFFFFFFF, w22=0xFFFFFFFF),
        Register(8, x23=0xFFFFFFFFFFFFFFFF, w23=0xFFFFFFFF),
        Register(8, x24=0xFFFFFFFFFFFFFFFF, w24=0xFFFFFFFF),
        Register(8, x25=0xFFFFFFFFFFFFFFFF, w25=0xFFFFFFFF),
        Register(8, x26=0xFFFFFFFFFFFFFFFF, w26=0xFFFFFFFF),
        Register(8, x27=0xFFFFFFFFFFFFFFFF, w27=0xFFFFFFFF),
        Register(8, x28=0xFFFFFFFFFFFFFFFF, w28=0xFFFFFFFF),
        Register(8, x29=0xFFFFFFFFFFFFFFFF, w29=0xFFFFFFFF),
        Register(8, x30=0xFFFFFFFFFFFFFFFF, w30=0xFFFFFFFF),
        Register(8, xzr=0xFFFFFFFFFFFFFFFF, wzr=0xFFFFFFFF),  # Zero register, must always equal zero

        Register(16, q0=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d0=0xFFFFFFFFFFFFFFFF, s0=0xFFFFFFFF, h0=0xFFFF, b0=0xFF),
        Register(16, q1=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d1=0xFFFFFFFFFFFFFFFF, s1=0xFFFFFFFF, h1=0xFFFF, b1=0xFF),
        Register(16, q2=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d2=0xFFFFFFFFFFFFFFFF, s2=0xFFFFFFFF, h2=0xFFFF, b2=0xFF),
        Register(16, q3=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d3=0xFFFFFFFFFFFFFFFF, s3=0xFFFFFFFF, h3=0xFFFF, b3=0xFF),
        Register(16, q4=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d4=0xFFFFFFFFFFFFFFFF, s4=0xFFFFFFFF, h4=0xFFFF, b4=0xFF),
        Register(16, q5=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d5=0xFFFFFFFFFFFFFFFF, s5=0xFFFFFFFF, h5=0xFFFF, b5=0xFF),
        Register(16, q6=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d6=0xFFFFFFFFFFFFFFFF, s6=0xFFFFFFFF, h6=0xFFFF, b6=0xFF),
        Register(16, q7=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d7=0xFFFFFFFFFFFFFFFF, s7=0xFFFFFFFF, h7=0xFFFF, b7=0xFF),
        Register(16, q8=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d8=0xFFFFFFFFFFFFFFFF, s8=0xFFFFFFFF, h8=0xFFFF, b8=0xFF),
        Register(16, q9=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d9=0xFFFFFFFFFFFFFFFF, s9=0xFFFFFFFF, h9=0xFFFF, b9=0xFF),
        Register(16, q10=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d10=0xFFFFFFFFFFFFFFFF, s10=0xFFFFFFFF, h10=0xFFFF, b10=0xFF),
        Register(16, q11=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d11=0xFFFFFFFFFFFFFFFF, s11=0xFFFFFFFF, h11=0xFFFF, b11=0xFF),
        Register(16, q12=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d12=0xFFFFFFFFFFFFFFFF, s12=0xFFFFFFFF, h12=0xFFFF, b12=0xFF),
        Register(16, q13=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d13=0xFFFFFFFFFFFFFFFF, s13=0xFFFFFFFF, h13=0xFFFF, b13=0xFF),
        Register(16, q14=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d14=0xFFFFFFFFFFFFFFFF, s14=0xFFFFFFFF, h14=0xFFFF, b14=0xFF),
        Register(16, q15=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d15=0xFFFFFFFFFFFFFFFF, s15=0xFFFFFFFF, h15=0xFFFF, b15=0xFF),
        Register(16, q16=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d16=0xFFFFFFFFFFFFFFFF, s16=0xFFFFFFFF, h16=0xFFFF, b16=0xFF),
        Register(16, q17=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d17=0xFFFFFFFFFFFFFFFF, s17=0xFFFFFFFF, h17=0xFFFF, b17=0xFF),
        Register(16, q18=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d18=0xFFFFFFFFFFFFFFFF, s18=0xFFFFFFFF, h18=0xFFFF, b18=0xFF),
        Register(16, q19=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d19=0xFFFFFFFFFFFFFFFF, s19=0xFFFFFFFF, h19=0xFFFF, b19=0xFF),
        Register(16, q20=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d20=0xFFFFFFFFFFFFFFFF, s20=0xFFFFFFFF, h20=0xFFFF, b20=0xFF),
        Register(16, q21=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d21=0xFFFFFFFFFFFFFFFF, s21=0xFFFFFFFF, h21=0xFFFF, b21=0xFF),
        Register(16, q22=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d22=0xFFFFFFFFFFFFFFFF, s22=0xFFFFFFFF, h22=0xFFFF, b22=0xFF),
        Register(16, q23=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d23=0xFFFFFFFFFFFFFFFF, s23=0xFFFFFFFF, h23=0xFFFF, b23=0xFF),
        Register(16, q24=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d24=0xFFFFFFFFFFFFFFFF, s24=0xFFFFFFFF, h24=0xFFFF, b24=0xFF),
        Register(16, q25=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d25=0xFFFFFFFFFFFFFFFF, s25=0xFFFFFFFF, h25=0xFFFF, b25=0xFF),
        Register(16, q26=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d26=0xFFFFFFFFFFFFFFFF, s26=0xFFFFFFFF, h26=0xFFFF, b26=0xFF),
        Register(16, q27=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d27=0xFFFFFFFFFFFFFFFF, s27=0xFFFFFFFF, h27=0xFFFF, b27=0xFF),
        Register(16, q28=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d28=0xFFFFFFFFFFFFFFFF, s28=0xFFFFFFFF, h28=0xFFFF, b28=0xFF),
        Register(16, q29=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d29=0xFFFFFFFFFFFFFFFF, s29=0xFFFFFFFF, h29=0xFFFF, b29=0xFF),
        Register(16, q30=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d30=0xFFFFFFFFFFFFFFFF, s30=0xFFFFFFFF, h30=0xFFFF, b30=0xFF),
        Register(16, q31=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, d31=0xFFFFFFFFFFFFFFFF, s31=0xFFFFFFFF, h31=0xFFFF, b31=0xFF),

        # CPSR register
        Register(
            4,
            n=0x80000000,  # negative flag
            z=0x40000000,  # zero flag
            c=0x20000000,  # carry flag
            v=0x10000000,  # overflow flag
            t=0x10,        # thumb mode flag
            f=0x20,        # FIQ disable
            i=0x40,        # IRQ disable
        ),
    ]

    # In 32-bit context, r13 is the same as sp, r14 is the same as lr and r15 is the same as pc.
    # So overwrite these 3 registers with new Register instances containing the extra alias.
    # In 64-bit context, sp and pc are separate registers, so add them now. lr is the same as r30
    # but since it is always referred to as r30 in assembly, no alias is created for it.
    if get_bits() == 64:
        registers.extend([
            Register(8, sp=0xFFFFFFFFFFFFFFFF, wsp=0xFFFFFFFF),
            Register(8, pc=0xFFFFFFFFFFFFFFFF)
        ])
    else:
        registers[13] = Register(8, x13=0xFFFFFFFFFFFFFFFF, w13=0xFFFFFFFF, r13=0xFFFFFFFF, sp=0xFFFFFFFF)
        registers[14] = Register(8, x14=0xFFFFFFFFFFFFFFFF, w14=0xFFFFFFFF, r14=0xFFFFFFFF, lr=0xFFFFFFFF)
        registers[15] = Register(8, x15=0xFFFFFFFFFFFFFFFF, w15=0xFFFFFFFF, r15=0xFFFFFFFF, pc=0xFFFFFFFF)

    return RegisterMap(registers)



"""
Core components that will be integrated as part of the root "kordesii" module.

NOTE: Importing "kordesii" is safe to do outside of IDA. Other modules, not so much.
"""

import atexit
import glob
import inspect
import io
import os
import logging
import re
import shutil
import subprocess
import sys
import threading

import pefile
from elftools.common.exceptions import ELFError
from elftools.elf import elffile

from kordesii import logutil

try:
    import idc

    in_ida = True
except ImportError:
    idc = None
    in_ida = False


logger = logging.getLogger(__name__)

# Special logger for IDA output stream.
# (using separate name to allow for filtering as desired.)
ida_logger = logging.getLogger("ida")


__all__ = [
    "decoder_entry",
    "called_from_framework",
    "in_ida",
    "find_ida",
    "run_ida",
    "is_64_bit",
    "append_string",
    "write_unique_file",
]

# Determines if the code was called from framework or directly from IDA
# TODO: Determine better way to do this.
called_from_framework = in_ida and "exit" in idc.ARGV

IDA_LOG_FILE = "ida_log.txt"
IDA_STRINGS_FILE = "ida_strings.txt"
DECODER_OUTPUT_DIR = "decoder_output_files"
RETURN_CODE_TIMEOUT = 20


def decoder_entry(main_func):
    """
    Main entry code that will trigger IDA script execution.

    Decorate your main function like so:

        @kordesii.script_entry
        def main():
            # ....

    WARNING: The decorated function MUST be at the end of the module.

    :param main_func: function to call
    """
    try:
        decoder_locals = inspect.stack(0)[1][0].f_locals
    except TypeError:
        return main_func
    if decoder_locals.get("__name__") == "__main__":
        if in_ida:
            # Setup logging.
            logutil.setup_logging()

            # Run main()
            idc.auto_wait()
            try:
                main_func()
            except Exception as e:
                # Catch all exceptions, otherwise we have to go fish it out of ida_log.txt
                logger.exception(e)
            finally:
                # Clear out the serializer so we can run multiple decoders in the same idb.
                # Must import here to avoid cyclic import.
                # FIXME: A proper fix for the serializer should be put in place.
                from kordesii import serialization

                serialization._serializers = {}
                # Exit if called from framework
                if called_from_framework:
                    idc.qexit(0)
        else:
            sys.exit("Script must be called from IDA or kordesii.")
            # TODO: We could possibly call run_ida() if outside.

    return main_func


def find_ida(is_64_bit=False):
    """
    Description:
        Find the highest version of IDA installed on the current (windows) system.

    Input:
        is_64_bit - Will return the path to the 64 bit IDA if True. Default False.

    Output:
        Return the absolute path to the highest version of IDA installed at the default location
        on the current windows system.

    Raises:
        IOError: If no installation of IDA could be found.
    """
    # Use user defined location if available.
    if "IDA_DIR" in os.environ:
        ida_dirs = [os.environ["IDA_DIR"]]
    else:
        # Find installed IDA paths.
        ida_dirs = glob.glob(r"C:\Program Files*\IDA *")

        # Sort by highest version.
        get_version = lambda path: path.rpartition(" ")[2]
        ida_dirs.sort(key=get_version, reverse=True)

    ida_exe_re = re.compile("idaq?64(\.exe)?$" if is_64_bit else "idaq?(\.exe)?$")

    # Find highest version with a ida.exe or idaq.exe in the directory.
    for ida_dir in ida_dirs:
        for filename in os.listdir(ida_dir):
            if ida_exe_re.match(filename):
                return os.path.abspath(os.path.join(ida_dir, filename))

    raise IOError("Unable to find IDA installation or executable. Please ensure IDA is installed.")


def is_64_bit(input_file):
    """
    Description:
        Do a quick check to see if it is an IDB/I64. Otherwise, attempt to determine if the
        file is 64bit based on the pe header. Note that the pe.close() prevents an mmap'd file from
        being left open indefinitely as the PE object doesn't seem to get garbage collected.
        Forcing garbage collection also corrects that issue.

    Input:
        input_file - The full path to the file in question

    Output:
        True if it is 64 bit. False if it is not or if pefile couldn't parse the header.
    """
    if input_file.endswith(".i64"):
        return True
    elif input_file.endswith(".idb"):
        return False

    # Get first bytes of file to check the file magic
    with open(input_file, "rb") as f:
        first_bytes = f.read(8)

    if first_bytes[0:2] == b"\x4D\x5A":
        # PE file type
        try:
            pe = pefile.PE(input_file, fast_load=True)
            result = pe.PE_TYPE == pefile.OPTIONAL_HEADER_MAGIC_PE_PLUS
            pe.close()
        except:
            result = False
    elif first_bytes[1:4] == b"\x45\x4C\x46":
        # elf file type
        try:
            with open(input_file, "rb") as f:
                elf = elffile.ELFFile(f)
                result = elf.get_machine_arch() in ["AArch64", "x64"]
        except ELFError:
            result = False
    elif first_bytes[0:4] == b"\xCE\xFA\xED\xFE":
        # 32 bit MACH-O executable
        result = False
    elif first_bytes[0:4] == b"\xCF\xFA\xED\xFE":
        # 64 bit MACH-O executable
        result = True
    else:
        result = False

    return result


def communicate(proc):
    """
    Drop the annoying swig memory leak warning. This is not a real concern for our purposes
    and only clutters the test results.
    """
    stdout, stderr = proc.communicate()
    ignore_str = b"swig/python detected a memory leak of type 'std::out_of_range *', no destructor found."
    if stdout:
        for out_line in stdout.splitlines(True):
            if out_line.strip() == ignore_str:
                continue
            sys.stdout.buffer.write(out_line)

    if stderr:
        for out_line in stderr.splitlines(True):
            if out_line.strip() == ignore_str:
                continue
            sys.stderr.buffer.write(out_line)


def run_ida(
    reporter,
    script_path,
    input_file,
    autonomous=True,
    ida_path=None,
    is_64bit=None,
    timeout=3600,
    log=False,
    cleanup_txt_files=True,
    cleanup_output_files=False,
    cleanup_idb_files=False,
):
    """
    Description:
        Call IDA given an input file and IDA script to run.

    Input:
        reporter - Kordesii reporter object
        script_path - Path to the IDA script (e.g. ida_configdumper.py).
        input_file - Path to the file for which an IDB should be built and analyzed.
        autonomous - When set to True, IDA will not display dialog boxes. Setting to False
                     is useful in debugging errors as the IDB will remain open barring an
                     explicit idc.qexit().
        ida_path - Full path to idaq.exe file to run IDA. If None, use find_ida() function to
                   determine the path.
        is_64bit - Boolean indicating if input file is 64 bit if True or 32 bit if False.
                   Parameter may also be None to be automatically determined based on the input file.
                   (this is the default)
        timeout - run_ida will wait <timeout> seconds before killing the process. Setting
                  timeout to 0 will disable the timeout. This is mostly useful for automation.
        log - Get IDA log contents. The output is equivalent to the IDA console output.
        cleanup_txt_files - Cleanup standard text files upon running ida like the log, strings, and debug files.
        cleanup_output_files - Cleanup any output files uniquely produced by the IDA script.
        cleanup_idbs - Cleanup any IDB and any IDB component files ('.til', '.nam', '.id0', '.id1', '.id2', '.id3')

    Output:
        Information will be added to the reporter object.
        Output files will exist after execution based on the cleanup parameters and the script itself.
    """
    # First find IDA executable to run
    if not ida_path:
        # Determine bitness of input file (or use user provided)
        if is_64bit is None:
            is_64bit = is_64_bit(input_file)
        ida_path = find_ida(is_64bit)

    # Setup some variables for files that may be output by the decoder and IDA
    base_dir = os.path.dirname(os.path.abspath(input_file))
    log_file_path = os.path.join(base_dir, IDA_LOG_FILE)
    strings_file_path = os.path.join(base_dir, IDA_STRINGS_FILE)
    output_dir_path = os.path.join(base_dir, DECODER_OUTPUT_DIR)

    # Cleanup any preexisting standard files to avoid overlap with previous decoder runs
    if os.path.exists(log_file_path):
        os.remove(log_file_path)
    if os.path.exists(strings_file_path):
        os.remove(strings_file_path)

    # Start the logging listener (if not already started)
    logutil.start_listener()
    assert logutil.listen_port

    # Setup the process to run the IDA decoder script
    command = [
        ida_path,
        "-P",
        '-S""{script_path}" {log_level} {log_port} exit"'.format(
            script_path=script_path, log_level=logging.root.getEffectiveLevel(), log_port=logutil.listen_port,
        ),
    ]
    if autonomous:
        command.append("-A")
    if log:
        command.append('-L"{}"'.format(log_file_path))

    command.append('"{}"'.format(input_file))
    command = " ".join(command)  # Doesn't work unless we convert to string!

    logger.debug("Running command: {}".format(command))
    process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=sys.platform != "win32")

    atexit.register(process.kill)

    # Wrap the call in a thread so we can time it out.
    thread = threading.Thread(target=communicate, args=(process,))
    thread.start()
    thread.join(timeout if timeout and timeout > 0 else None)  # Block on timeout of 0
    if thread.is_alive():  # This will only be true if timeout != 0
        logger.info("Killing IDA process: exceeded timeout of " + str(timeout))
        process.kill()
        thread.join()
        return

    # IDA script completed
    # Make a note of the return code to have the information
    if process.returncode == 0:
        logger.info("IDA return code = {}".format(process.returncode))
    else:
        logger.error("IDA return code = {}".format(process.returncode))

    # Ingest any debug information output by the script
    # TODO: Determine if/how we can pull the console output as it's produced.
    #   If possible, we could feed the output as debug logs instead of treating them differently.
    if log and os.path.isfile(log_file_path):
        with io.open(log_file_path, "r", encoding="utf-8", errors="replace") as f:
            reporter.ida_log = f.read()

            # Also throw logs to debug.
            f.seek(0)
            for line in f.readlines():
                ida_logger.debug(line.rstrip("\r\n"))

    # Ingest any strings output by the script
    if os.path.isfile(strings_file_path):
        with open(strings_file_path, "r") as f:
            # NOTE: We can't sort and dedup because other parsers may depend on their order.
            strings = [entry.rstrip("\r\n") for entry in f.readlines()]
            # print("strings: " + repr(strings))
        for string in strings:
            try:
                reporter.add_string(string.encode("latin1").decode("unicode-escape"))
            except Exception as e:
                logger.error("Bad string {!r}: {}".format(string, e))

    # Ingest any files output by the script
    if os.path.exists(output_dir_path):
        for root, dirs, files in os.walk(output_dir_path):
            for file in files:
                file_path = os.path.join(root, file)
                with open(file_path, "rb") as fo:
                    file_data = fo.read()
                reporter.add_output_file(file, file_data)

    # Perform cleanup as specified in the parameters
    if cleanup_txt_files:
        if os.path.exists(log_file_path):
            os.remove(log_file_path)
        if os.path.exists(strings_file_path):
            os.remove(strings_file_path)
    if cleanup_output_files:
        if os.path.exists(output_dir_path):
            shutil.rmtree(output_dir_path)
    if cleanup_idb_files:
        _remove_idbs(input_file)


def _remove_idbs(input_file):
    """
    Description:
        Remove all IDB/I64 files and their components from the specified directory.
        Assumes the files aren't still open somewhere.

    Input:
        input_file - The file used to create the IDB we want to remove or the IDB itself.
    """
    try:
        if os.path.exists(input_file + ".idb"):
            os.remove(input_file + ".idb")
        elif os.path.exists(input_file + ".i64"):
            os.remove(input_file + ".i64")
    except OSError:
        logger.warning("Unable to remove " + input_file)

    # The order of the extensions here is important.
    for ext in (".til", ".nam", ".id0", ".id1", ".id2", ".id3"):
        try:
            os.remove(input_file + ext)
        except OSError:
            break  # The file didn't exist. Since the extensions are ordered, we can assume none
            # of the others do either.


# TODO: Use the serializer instead of this IDA_STRINGS_FILE
def append_string(string):
    """
    Append decoded string to file for access outside of IDA.

    :raises ValueError: If run outside of IDA.
    """
    if not in_ida:
        raise ValueError("This function can only be run within IDA.")

    try:
        # Make sure string is unicode escaped before writing!
        if not isinstance(string, str):
            string = string.decode("latin1")
        string = string.encode("unicode-escape")
        with open(IDA_STRINGS_FILE, "ab") as f:
            f.write(b"".join([string, b"\n"]))
    except Exception as e:
        logger.error("Error writing string to %s: %s" % (IDA_STRINGS_FILE, str(e)))


def write_unique_file(filename, data):
    """
    Some IDA scripts will output files in addition to strings and debug messages.
    This function will append the provided data to the specified file name in
    a designated directory for access later.

    :raises ValueError: If run outside of IDA.
    """
    if not in_ida:
        raise ValueError("This function can only run within IDA.")

    if not os.path.exists(DECODER_OUTPUT_DIR):
        os.makedirs(DECODER_OUTPUT_DIR)

    filepath = os.path.join(DECODER_OUTPUT_DIR, filename)

    try:
        with open(filepath, "wb") as f:
            f.write(data)
        logger.info("Wrote file: {}".format(filepath))
    except Exception as e:
        logger.error("Error writing data to %s: %s" % (filepath, str(e)))



"""
IDA Pyro4 server used to run IDAPython remotely.
"""
import functools
import inspect
import importlib
import io
import logging
import pickle
import sys
import threading
import traceback
import types
import warnings

import dill
import Pyro4

import kordesii


daemon_stop = threading.Event()
_close_ida = True


logger = kordesii.get_logger()


# The following list contains all the modules to be proxied into IDA.
proxied_modules = [
    "idc",
    "idaapi",
    "idautils",
    "ida_allins",
    "ida_auto",
    "ida_bytes",
    "ida_dbg",
    "ida_diskio",
    "ida_entry",
    "ida_enum",
    "ida_expr",
    "ida_fixup",
    "ida_fpro",
    "ida_frame",
    "ida_funcs",
    "ida_gdl",
    "ida_graph",
    "ida_hexrays",
    "ida_ida",
    "ida_idaapi",
    "ida_idc",
    "ida_idd",
    "ida_idp",
    "ida_kernwin",
    "ida_lines",
    "ida_loader",
    "ida_moves",
    "ida_nalt",
    "ida_name",
    "ida_netnode",
    "ida_offset",
    "ida_pro",
    "ida_problems",
    "ida_range",
    "ida_registry",
    "ida_search",
    "ida_segment",
    "ida_segregs",
    "ida_strlist",
    "ida_struct",
    "ida_tryblks",
    "ida_typeinf",
    "ida_ua",
    "ida_xref",
    # Add kordesii modules that will work proxied.
    # (ie. modules that have functions only and no exposed classes)
    "kordesii.utils.utils",
    "kordesii.utils.segments",
]


class IDAProxy(object):
    r"""
    This is the proxy class. This class is made accessible to client
    (kordesii) through Pyro4.

    Nearly all of the methods are `staticmethod`\ s or properties.
    This class exists to be a proxy for the IDA-specific functions and methods.
    These are added later with the :meth:`~add_module` function.
    """

    # expose at least one method so Pyro doesn't have a fit.
    @staticmethod
    def testing_method():
        return "HI!!"


class IDAModule(object):
    """
    This is a simple carrier class to pass the object id over to ida_client.
    When ida_client sees this class, it will replace it wih a proxy of the specified object id.
    """
    def __init__(self, object_id):
        self.object_id = object_id


class IDAClass(IDAModule):
    pass


def build_proxy_property(prop):
    """
    Builds a very simple proxy for a property.

    This is required in order add a property to the proxy class.
    """

    # noinspection PyUnusedLocal
    def _proxy(self):
        return prop

    return property(_proxy)


def build_proxy_submodule(module):
    """
    Creates a simple dummy class (IDAModule) containing the object id so ida_client
    can pick this up and initialize the proxy on its side.
    """
    def _proxy(self):
        if module.__name__ in proxied_modules:
            return IDAModule(module.__name__)
        return None
    return property(_proxy)


def build_proxy_class(klass, klass_name):
    """
    Creates a simple dummy class (IDAClass) containing the object id so ida_client can
    pick this up and initialize the proxy on its side.
    This is necessary since Pyro4's autoproxy feature causes too many issues.
    """
    object_id = ".".join([klass.__module__, klass_name])
    ida_class = IDAClass(object_id)

    def _proxy(self):
        if object_id not in self._pyroDaemon.objectsById:
            proxied_klass = _generate_proxy_class(klass, klass_name, ignore_classes=True)
            self._pyroDaemon.register(proxied_klass, objectId=object_id)
        return ida_class

    return property(_proxy)


def add_module(mod, klass, ignore_classes=False):
    """
    Adds as much of a module as possible to the given class.

    Sub-modules are not added.

    Methods are proxied and added as static methods. Properties
    are similarly proxied and added as properties.

    :param mod: Module to add to the given class.
    :param klass: Class to add the module's functions and attributes to.
    """
    for member_name in dir(mod):
        member = getattr(mod, member_name)

        # Some hackery needed because Pyro4 doesn't like private methods and attributes.
        if member_name.startswith("_"):
            member_name = "priv" + member_name

        if hasattr(klass, member_name):
            continue

        # Wrap functions and class initializations.
        if inspect.isroutine(member):
            setattr(klass, member_name, staticmethod(member))
        elif isinstance(member, type):
            # Ignore setting classes ontop classes.
            if isinstance(mod, type) or ignore_classes:
                continue
            setattr(klass, member_name, build_proxy_class(member, member_name))
        elif isinstance(member, types.ModuleType):
            setattr(klass, member_name, build_proxy_submodule(member))
        else:
            setattr(klass, member_name, build_proxy_property(member))


class MainProxy(object):
    """
    This class holds the functions and attributes available in the main proxy.

    This is used to act as a controller for the external IDA client.
    This allows us to stop the IDA server as well as run generic functions
    that were originally unexposed.
    """

    @staticmethod
    def stop_daemon(close_ida=True):
        """
        Terminate the Pyro4 daemon and (optionally) teardown IDA

        :param bool close_ida: If IDA should be closed
        """
        global _close_ida

        daemon_stop.set()
        if not close_ida:
            _close_ida = False

    @staticmethod
    def run_func(path, _globals, func, *args, **kwargs):
        """
        Runs any generic function.

        :param path: The sys.path from the caller.
            This helps to discover functions not in a package (ala, scripts)
        :param func: The pickled function to run
        :param _globals: Exposed global functions marked as run_in_ida.
        :param args: positional arguments
        :param kwargs: keyword arguments.
        :return: Return value of function.
        """
        orig_path = sys.path[:]
        orig_stdout = sys.stdout
        orig_stderr = sys.stderr

        try:
            # update sys.paths, so we can find the function
            sys.path += path
            # redirect stdout and stderr so we can bring it over
            sys.stdout = io.StringIO()
            sys.stderr = io.StringIO()

            func = dill.loads(func)
            _globals = dill.loads(_globals)

            try:
                # Apply exposed global functions.
                globals().update(_globals)

                # Now run the function.
                ret = func(*args, **kwargs)
            except Exception as e:
                # Pass any exceptions thrown as the return value
                # so we can reraise it externally.
                # Attach the traceback, so it can be printed out externally.
                e.ida_traceback = traceback.format_exception(None, e, e.__traceback__)
                ret = e

            return ret, sys.stdout.getvalue(), sys.stderr.getvalue()
        finally:
            sys.path = orig_path
            sys.stdout = orig_stdout
            sys.stderr = orig_stderr

    @staticmethod
    def ping():
        """Check connection is working."""
        return True


memo = {}


def _generate_proxy_class(module_or_class, name, ignore_classes=False):
    """Generates a safe proxyable class from given module or class."""
    if module_or_class not in memo:
        # Pyro4 doesn't allow private module names to be exposed, so a little hackery is in order.
        klass = type(name.strip("_"), (IDAProxy,), {})
        add_module(module_or_class, klass, ignore_classes=ignore_classes)
        Pyro4.expose(klass)
        memo[module_or_class] = klass
        return klass
    else:
        return memo[module_or_class]


def _register(daemon):
    """
    Sets up IDA/kordesii modules to be expose to given Pyro4 daemon.

    :returns: uri information returned by Pyro4
    :raises: An exception could be raised if we fail to import a module.
    """
    for module_name in proxied_modules:
        logging.debug("registering {}".format(module_name))
        module = importlib.import_module(module_name)
        klass = _generate_proxy_class(module, module_name)
        daemon.register(klass, objectId=module_name)

    # Now expose the main Proxy controller.
    Pyro4.expose(MainProxy)
    uri = daemon.register(MainProxy, objectId="main")

    return uri


def _send_result(result):
    """
    Write out a file telling client the port that was used or the exception that has occurred.
    (this also doubles as a way for the client to know we are ready)
    """
    import idc
    with open(idc.ARGV[3], "wb") as f:
        f.write(dill.dumps(result))


def main():
    # Don't support the legacy api.
    sys.modules["__main__"].IDAPYTHON_COMPAT_695_API = False

    import idc

    kordesii.setup_logging()

    try:
        Pyro4.config.SERVERTYPE = "multiplex"
        Pyro4.config.FLAME_ENABLED = "True"
        Pyro4.config.SERIALIZERS_ACCEPTED = {"dill"}

        logger.debug("Starting daemon...")
        daemon = Pyro4.Daemon(host="localhost")
        warnings.simplefilter("ignore")

        uri = _register(daemon)

        logger.info("Listening on {}".format(uri))
        # Send port back to the client.
        _send_result(uri.port)

        # Start listener
        idc.auto_wait()
        daemon.requestLoop(loopCondition=lambda: not daemon_stop.is_set())

        if _close_ida:
            idc.qexit(0)

    except Exception as e:
        # Send exception back to the client, so they can raise it outside of IDA.
        _send_result(e)


if __name__ == "__main__" and kordesii.in_ida:
    logger.info("Starting")
    main()



"""
Common standard C library builtin functions.
"""

import logging
import re

from ... import utils
from ... import constants
from ...call_hooks import builtin_func

logger = logging.getLogger(__name__)


@builtin_func("_alloca")
@builtin_func("_malloca")
def _alloca(cpu_context, func_name, func_args):
    """
    Allocates stack space

    NOTE: Our memory controller will automatically allocate pages as it writes.
    Therefore, we don't actually need to allocate anything, just provide them with
    a pointer.
    """
    (size,) = func_args
    if size:
        size = utils.align_page_up(size)
        logger.debug("Allocating %d bytes of stack space.", size)
        # Returns a pointer to our allocated stack space.
        return cpu_context.sp - size


@builtin_func("malloc")
@builtin_func("calloc")
def malloc(cpu_context, func_name, func_args):
    """
    Allocates memory block
    """
    if func_name == "calloc":
        num, size = func_args
        size = num * size
    else:
        (size,) = func_args
    if size:
        logger.debug("Allocating %d bytes of memory.", size)
        return cpu_context.mem_alloc(size)


@builtin_func
def new(cpu_context, func_name, func_args):
    """
    Allocates memory block
    """
    size = func_args[0]
    # If we have a second argument it's either a nothrow_value or a ptr to where they
    # want us to allocate the space.
    # Just assume its a ptr if it's not 0 since the emulator should not produce a nothrow_value correctly
    # and therefore be 0.
    # TODO: Perhaps provide more context to make a more informed decision.
    try:
        ptr = func_args[1]
    except IndexError:
        ptr = 0
    if size:
        if ptr:
            logger.debug("Mapping %d bytes into 0x%X", size, ptr)
            # If a pointer was provided, just give it back to them without doing anything.
            # Our memory controller will automatically allocate pages as they are being written.
            return ptr

        logger.debug("Allocating %d bytes of memory.", size)
        return cpu_context.mem_alloc(size)


@builtin_func
def realloc(cpu_context, func_name, func_args):
    """
    Reallocates memory block
    """
    ptr, size = func_args
    if ptr and size:
        logger.debug("Reallocating 0x%X with %d bytes.", ptr, size)
        return cpu_context.mem_realloc(ptr, size)


@builtin_func
def memchr(cpu_context, func_name, func_args):
    """
    Locate character in block of memory.

    Searches within the first num bytes of the block of memory pointed by data_ptr
    for the first occurrence of value (interpreted as an unsigned char) and returns a pointer
    to it.
    """
    data_ptr, value, num = func_args

    ptr = cpu_context.memory.find(value, start=data_ptr, end=data_ptr + num)
    if ptr == -1:
        return 0

    return ptr


@builtin_func("memmove")
@builtin_func("memmove_s")
@builtin_func("wmemmove")
@builtin_func("wmemmove_s")
@builtin_func("memcpy")
@builtin_func("memcpy_s")
@builtin_func("wmemcpy")
@builtin_func("wmemcpy_s")
def memcpy(cpu_context, func_name, func_args):
    """
    Copies count characters from src to dst.
    """
    secure = func_name.endswith("_s")
    wide = func_name.startswith("w")
    if secure:
        dst, dst_size, src, count = func_args
        count = min(dst_size, count)
    else:
        dst, src, count = func_args

    if wide:
        count *= 2

    if dst and src:
        logger.debug("Copying %d bytes from 0x%X to 0x%X", count, src, dst)
        cpu_context.mem_copy(src, dst, count)
        return 0 if secure else dst


@builtin_func
def memset(cpu_context, func_name, func_args):
    """
    Writes count number of ch characters to dst.
    """
    dst, ch, count = func_args
    if dst:
        logger.debug("Writing %r * %d to 0x%X", chr(ch), count, dst)
        cpu_context.mem_write(dst, bytes([ch]) * count)
    return dst


@builtin_func
def memcmp(cpu_context, func_name, func_args):
    lhs, rhs, count = func_args
    if lhs and rhs:
        logger.debug("Comparing the first %d bytes in 0x%X with 0x%X", count, lhs, rhs)
        left = cpu_context.mem_read(lhs, count)
        right = cpu_context.mem_read(rhs, count)
        if left < right:
            return -1

        if left > right:
            return 1

        return 0


@builtin_func("strcat")
@builtin_func("strcat_s")
@builtin_func("wcscat")
@builtin_func("wcscat_s")
@builtin_func("lstrcatA")
@builtin_func("lstrcatW")
def strcat(cpu_context, func_name, func_args):
    """
    Concatenates at most count characters from src, stopping if the null character
    is found, to the end of the null-terminated byte string pointed to by dst.
    The src[0] replaces the null terminator at the end of dst.
    """
    secure = func_name.endswith("_s")
    wide = func_name.startswith("w") or func_name.endswith("W")
    if secure:
        dst, dst_size, src = func_args
        if wide:
            dst_size *= 2
    else:
        dst, src = func_args
        dst_size = None

    if dst and src:
        logger.debug("Concatenating c string in 0x%X to c string in 0x%X", src, dst)
        append_str = cpu_context.read_data(src, data_type=constants.WIDE_STRING if wide else constants.STRING)
        dest_str = cpu_context.read_data(dst, data_type=constants.WIDE_STRING if wide else constants.STRING)
        null_offset = dst + len(dest_str)
        if dst_size is not None:
            dst_size -= null_offset - dst
            append_str = append_str[:dst_size]
        terminator = b"\0\0" if wide else b"\0"
        cpu_context.mem_write(null_offset, append_str + terminator)
        return 0 if secure else dst


@builtin_func("strncat")
@builtin_func("strncat_s")
@builtin_func("wcsncat")
@builtin_func("wcsncat_s")
def strncat(cpu_context, func_name, func_args):
    """
    Concatenates at most count characters from src, stopping if the null character
    is found, to the end of the null-terminated byte string pointed to by dst.
    The src[0] replaces the null terminator at the end of dst.
    """
    secure = func_name.endswith("_s")
    wide = func_name.startswith("w")
    if secure:
        dst, dst_size, src, count = func_args
        if wide:
            dst_size *= 2
    else:
        dst, src, count = func_args
        dst_size = None
    if wide:
        count *= 2

    if dst and src:
        logger.debug("Concatenating c string in 0x%X to c string in 0x%X", src, dst)
        append_str = cpu_context.read_data(src, data_type=constants.WIDE_STRING if wide else constants.STRING)[:count]
        dest_str = cpu_context.read_data(dst, data_type=constants.WIDE_STRING if wide else constants.STRING)
        null_offset = dst + len(dest_str)
        if dst_size is not None:
            dst_size -= null_offset - dst
            append_str = append_str[:dst_size]
        terminator = b"\0\0" if wide else b"\0"
        cpu_context.mem_write(null_offset, append_str + terminator)
        return 0 if secure else dst


@builtin_func
def strpbrk(cpu_context, func_name, func_args):
    """
    Locate characters in string.

    Returns a pointer to the first occurrence of str1 of any of the characters
    that are part of str2, or a null pointer if there are no matches.
    """
    str1_ptr, str2_ptr = func_args
    str1 = cpu_context.read_data(str1_ptr)
    str2 = cpu_context.read_data(str2_ptr)

    for offset, ch in enumerate(str1):
        if ch in str2:
            return str1_ptr + offset
    return 0


@builtin_func("strchr")
@builtin_func("strrchr")
def strchr(cpu_context, func_name, func_args):
    """
    Locate first or last occurrence of character in string.
    """
    string_ptr, character = func_args
    string = cpu_context.read_data(string_ptr)

    if func_name == "strchr":
        offset = string.find(character)
    else:
        offset = string.rfind(character)

    if offset == -1:
        return 0

    return string_ptr + offset


@builtin_func("strcpy")
@builtin_func("strcpy_s")
@builtin_func("wcscpy")
@builtin_func("wcscpy_s")
@builtin_func("lstrcpyA")
@builtin_func("lstrcpyW")
def strcpy(cpu_context, func_name, func_args):
    """
    Copies the null-terminated byte string pointed to by src, including the null terminator, to dst.
    """
    secure = func_name.endswith("_s")
    wide = func_name.startswith("w") or func_name.endswith("W")
    if secure:
        dst, dst_size, src = func_args
        if wide:
            dst_size *= 2
    else:
        dst, src = func_args
        dst_size = None

    if dst and src:
        logger.debug("Copying c string in 0x%X to 0x%X", src, dst)
        terminator = b"\0\0" if wide else b"\0"
        src_str = cpu_context.read_data(src, data_type=constants.WIDE_STRING if wide else constants.STRING) + terminator
        size = len(src_str)
        if dst_size is not None:
            size = min(dst_size, size)  # limit to dst_size if secure.
        cpu_context.mem_copy(src, dst, size)
        return 0 if secure else dst


@builtin_func("strncpy")
@builtin_func("strncpy_s")
@builtin_func("wcsncpy")
@builtin_func("wcsncpy_s")
@builtin_func("lstrcpynA")
@builtin_func("lstrcpynW")
def strncpy(cpu_context, func_name, func_args):
    """
    Copies at most count characters of src (including the terminating null character) to dst
    but not any of the characters that follow the null character.
    For non-secure versions, if count is not reached, additional null characters are written to dst.
    """
    secure = func_name.endswith("_s")
    wide = func_name.startswith("w") or func_name.endswith("W")
    if secure:
        # Secure version has an extra dst_size argument used for validation, but we don't care to
        # throw errors if we don't have to!
        dst, _, src, count = func_args
    else:
        dst, src, count = func_args
    if wide:
        count *= 2

    if dst and src:
        logger.debug("Copying c string in 0x%X to 0x%X", src, dst)
        terminator = b"\0\0" if wide else b"\0"
        src_str = cpu_context.read_data(src, data_type=constants.WIDE_STRING if wide else constants.STRING) + terminator
        size = min(count, len(src_str))
        cpu_context.mem_copy(src, dst, size)
        # Non-secure version also pads the rest with null characters to reach count size
        if not secure and size < count:
            # As a safety check, we shouldn't be writing more than about 0x1000 of padding.
            delta = count - size
            if delta > 0x1000:
                logger.warning(
                    "Attempted to write %d bytes of padding. Ignoring request and using %d bytes "
                    "of padding instead.", delta, 0x1000)
                delta = 0x1000
            cpu_context.mem_write(src + size, b"\0" * delta)
        return 0 if secure else dst


@builtin_func
def strdup(cpu_context, func_name, func_args):
    """
    Returns a pointer to a null-terminated byte string, which is a duplicate of the string
    pointing to by str.
    """
    (str_ptr,) = func_args
    if str_ptr:
        logger.debug("Copying c string in 0x%X to new pointer.", str_ptr)
        null_offset = cpu_context.memory.find(b"\0", start=str_ptr)
        size = null_offset - str_ptr
        # create new pointer
        new_ptr = cpu_context.mem_alloc(size + 1)
        # Copy at most size bytes of data then add the null terminator.
        cpu_context.mem_copy(str_ptr, new_ptr, size)
        cpu_context.mem_write(new_ptr + size, b"\0")
        return new_ptr


@builtin_func
def strndup(cpu_context, func_name, func_args):
    """
    Returns a pointer to a null-terminated byte string, which contains copies
    of at most size bytes from the string pointed to by str.
    """
    str_ptr, size = func_args
    if str_ptr:
        logger.debug("Copying %d bytes of c string in 0x%X to new pointer.", size, str_ptr)
        null_offset = cpu_context.memory.find(b"\0", start=str_ptr)
        size = min(size, null_offset - str_ptr)
        # create new pointer
        new_ptr = cpu_context.mem_alloc(size + 1)
        # Copy at most size bytes of data then add the null terminator.
        cpu_context.mem_copy(str_ptr, new_ptr, size)
        cpu_context.mem_write(new_ptr + size, b"\0")
        return new_ptr


@builtin_func("strlen")
@builtin_func("strnlen_s")
@builtin_func("wcslen")
@builtin_func("wcsnlen_s")
@builtin_func("lstrlenA")
@builtin_func("lstrlenW")
def strlen(cpu_context, func_name, func_args):
    """
    Returns the length of the given null-terminated byte string.
    """
    secure = func_name.endswith("_s")
    wide = func_name.startswith("w") or func_name.endswith("W")
    if secure:
        str_ptr, strsz = func_args
    else:
        (str_ptr,) = func_args
        strsz = None

    if str_ptr:
        logger.debug("Getting length of c string in 0x%X", str_ptr)
        sstr = cpu_context.read_data(str_ptr, data_type=constants.WIDE_STRING if wide else constants.STRING)
        size = len(sstr)
        if wide:
            size /= 2
        # For secure version, strsz is returned if the terminator was not found
        # in the first strsz characters.
        if strsz is not None and size > strsz:
            return strsz
        return size


@builtin_func
def strstr(cpu_context, func_name, func_args):
    """
    Locate substring.

    Returns a pointer to the first occurrence of str2 in str1, or a null pointer if str2 is not part of str1.
    """
    str1_ptr, str2_ptr = func_args
    str1 = cpu_context.read_data(str1_ptr)
    str2 = cpu_context.read_data(str2_ptr)

    offset = str1.find(str2)
    if offset == -1:
        return 0

    return str1_ptr + offset


def _format_string(ctx, fmt, required_args, string_type=constants.STRING):
    """
    Handles formatting the string with the function arguments based on the format.

    :param ctx: cpu_context object
    :param fmt: format string
    :param required_args: num of required arguments for the particular format function to skip
    """
    # Format using best attempt here.  Basically, locate all the format specifiers, and convert them to a python
    # supported format string.  For each format string, extract the appropriate data from the context, and append it to
    # the values list.
    fmt_val_re = re.compile(br"""
    %                           # start with percent character
    [-+ #0]{0,1}                # optional flag character
    (\*|[0-9]{1,}){0,}          # optional width specifier, though mutually exclusive (either a number or *, not both)
    ((\.[0-9]{1,})|\.\*){0,}    # optional precision specifier, mutually exclusive
    [diuoxXfFeEgGaAcspn]        # format type
    """, re.VERBOSE)

    # NOTE: findall() produces empty results so we need to use finditer()
    fmt_vals = [match.group() for match in fmt_val_re.finditer(fmt)]
    logger.debug("Format vals: %r", fmt_vals)

    # Re-pull function arguments with correct number of arguments.
    func_sig = ctx.get_function_signature()
    func_sig.arg_types = func_sig.arg_types[:required_args]
    # For an unknown reason, int is not always being read as a QWORD on 64-bit, so this line
    # forces the issue to ensure pointer addresses aren't being truncated to 32 bits
    types = "QWORD" if utils.get_bits() == 64 else "DWORD"
    func_sig.arg_types += ((types,) * len(fmt_vals))
    func_args = [arg.value for arg in func_sig.args]

    format_vals = []
    arg_pos = required_args  # skip destination and format string
    for match in fmt_vals:
        if b'*' in match:
            # Indicates that one of the parameters is a width, which must be pulled and added to the list first
            format_vals.append(func_args[arg_pos])
            arg_pos += 1

        if match.endswith(b'c'):  # character (will this be the value or a read from the context???
            arg_val = func_args[arg_pos]
            if arg_val <= 0xFF:  # assume that the argument contains the character
                format_vals.append(arg_val)
            else:   # assume it's a pointer that must be dereferenced
                format_vals.append(ctx.read_data(arg_val, size=1))

        elif match.endswith(b's'):  # string value, should be a pointer
            _arg = ctx.read_data(func_args[arg_pos], data_type=string_type)
            if not len(_arg):   # If the argument isn't set during parsing, preserve the formatting
                logger.debug("Pulled 0 byte format string, reverting")
                _arg = b"%s"
            format_vals.append(_arg)

        else:   # all other numerical types???
            format_vals.append(func_args[arg_pos])

        arg_pos += 1

    result = fmt % tuple(format_vals)
    return result


@builtin_func
def sprintf(ctx, func_name, func_args):
    """
    Format a string based on provided format string and parameters.

    For sprintf, there's no way to know up front how many args are needed, but there should always be at least
    2 (destination and format).  We can use the format string to determine how many arguments we need by
    counting the format specifiers.
    """
    # Almost guaranteed to get the incorrect number of args.  So obtain the format string and count the number of
    # format specifiers to determine how many args we need, not including the first 2
    if len(func_args) < 2:   # Ensure that there are at least 2 arguments, dest and format
        # Need to try to get at least 2 arguments...
        func_args = ctx.get_function_args(num_args=2)

    dest = func_args[0]
    fmt = ctx.read_data(func_args[1])
    logger.debug("Format string: %s", fmt)
    result = _format_string(ctx, fmt, 2)
    logger.debug("Writing formatted value %s to 0x%X", result, dest)
    ctx.mem_write(dest, result + b'\0')
    return len(result)


@builtin_func("snprintf")
@builtin_func("sprintf_s")
@builtin_func("swprintf_s")
def snprintf(ctx, func_name, func_args):
    """
    int snprintf (char *s, size_t n, const char *format, ...);

    Format a string using the provided format string and values, truncated if necessary to length n.
    """
    wide = func_name.startswith("sw")
    string_type = constants.WIDE_STRING if wide else constants.STRING
    if len(func_args) < 3:
        func_args = ctx.get_function_args(num_args=3)

    dest, n = func_args[:2]
    fmt = ctx.read_data(func_args[2], data_type=string_type)
    logger.debug("Format string: %s", fmt)
    result = _format_string(ctx, fmt, 3, string_type)
    logger.debug("Writing formatted value %s to 0x%X", result[:n - 1], dest)
    ctx.mem_write(dest, result[:n - 1] + b'\0')
    return len(result)


@builtin_func
def printf(ctx, func_name, func_args):
    """
    int printf (const char * format, ...);

    Writes the C string pointed by **format** to the standard output (stdout).
    If **format** includes format specifiers (subsequences beginning with %), the
    additional arguments following **format** are formatted and inserted in the resulting string replacing their
    respective specifiers.
    """
    if len(func_args) < 1:
        func_args = ctx.get_function_arg_values(num_args=1)

    fmt = ctx.read_data(func_args[0])
    logger.debug("Format string: %s", fmt)
    result = _format_string(ctx, fmt, 1)
    logger.debug("Writing formatted value %s to stdout", result)
    ctx.stdout += result.decode()
    return len(result)



"""
Utilties for working with functions.
"""

import collections
import logging
import warnings
from typing import Iterable, Tuple

import ida_bytes
import ida_idp
import ida_name
import ida_ua
import idaapi
import idautils
import idc

from kordesii.utils.flowchart import Flowchart
from kordesii.utils.function_creator import create_function_precise
from kordesii import utils

logger = logging.getLogger(__name__)


class Function(object):
    """
    Object that wraps the func_t object created by IDA.
    This object greatly expands the functionality for interfacing with a function.
    """

    def __init__(self, ea, identifier=None, create_if_not_exists=True):
        """

        :param ea: An address within the function.
        :param identifier: **DEPRECATED** Some custom unique identifier for this function.
            Historically this was the id of the YARA rule that hit this function.
        :param create_if_not_exists: If True, the function_creator will be used to attempt
            to create function if not defined by IDA.

        :raises AttributeError: If no function exists at the given address.
        """
        self.origin_ea = ea
        self._identifier = identifier
        self.function_obj = idaapi.get_func(ea)
        if not self.function_obj:
            if create_if_not_exists and utils.is_x86_64():  # Function creator only supports x86
                if create_function_precise(ea, False):
                    self.function_obj = idaapi.get_func(ea)
                    logger.debug(
                        "Created function at 0x%X" % self.function_obj.start_ea
                    )
                else:
                    raise AttributeError("No function at 0x%X" % ea)
            else:
                raise AttributeError("No function at 0x%X" % ea)
        if self.function_obj:
            self.start_ea = self.function_obj.start_ea
            self.end_ea = self.function_obj.end_ea
        self.name = idaapi.get_func_name(self.function_obj.start_ea)
        self.xrefs_to = [
            ref.frm
            for ref in idautils.XrefsTo(self.function_obj.start_ea)
            if idaapi.get_func_name(ref.frm) != self.name
        ]
        self.xref_count = len(self.xrefs_to)
        self._flowchart = None
        self._api_calls = None

    @property
    def identifier(self):
        warnings.warn("The identifier property in Function is deprecated.", DeprecationWarning)
        return self._identifier

    @identifier.setter
    def identifier(self, value):
        warnings.warn("The identifier property in Function is deprecated.", DeprecationWarning)
        self._identifier = value

    @classmethod
    def from_name(cls, func_name: str, ignore_underscore: bool = False) -> "Function":
        """
        Factory method for obtaining Function by name.

        :param str func_name: Name of function to obtain
        :param bool ignore_underscore: Whether to ignore underscores in function name.
            (Will return the first found function if enabled.)

        :return: Function object
        :raises ValueError: If function name was not found.
        """
        for ea in idautils.Functions():
            _func_name = idc.get_func_name(ea)
            if ignore_underscore:
                _func_name = _func_name.strip("_")
            if func_name == _func_name:
                return cls(ea)
        raise ValueError("Unable to find function with name: {}".format(func_name))

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()

    def __hash__(self):
        return self.__repr__().__hash__()

    def __str__(self):
        return "%s 0x%X - 0x%X" % (
            self.name,
            self.function_obj.start_ea,
            self.function_obj.end_ea,
        )

    def __repr__(self):
        return "<Function : {}() : {:#08x} - {:#08x}>".format(
            self.name, self.start_ea, self.end_ea
        )

    def __contains__(self, ea: int):
        """Tests if ea is within function."""
        # NOTE: We can't just test if it's between start_ea and end_ea because the function might have
        # fragmented function chunks.
        try:
            func = idaapi.get_func(ea)
            if not func:
                return False
        except AttributeError:
            return False
        return func.start_ea == self.start_ea

    def heads(self, start: int = None, reverse: bool = False, dfs: bool = False) -> Iterable[int]:
        """
        Iterates all the heads for the given function.

        :param start: Start address (defaults to start_ea or end_ea)
        :param reverse:  Direction to iterate
        :param bool dfs: If true, traversal of blocks will be depth-first.
            If false, traversal will be breadth-first.

        :yields: Address of head
        """
        if not self._flowchart:
            self._flowchart = Flowchart(self.start_ea)
        if not start:
            start = self.end_ea if reverse else self.start_ea

        yield from self._flowchart.heads(start, reverse=reverse, dfs=dfs)

    def rename(self, new_name: str):
        """
        Attempts to apply new_name to the object at <ea>. If more than one object starts at <ea>, the
        largest object will be renamed. If that name already exists, let IDA resolve the collision
        and then return that name. If new_name is "", reset the name to IDA's default.

        :param str new_name: The desired new name for the function.

        :return str: The name that ended up getting set (unless no name was set, then return None).
        """
        if new_name == "":
            if idaapi.set_name(self.start_ea, new_name):
                return idaapi.get_name(self.function_obj.start_ea)
            else:
                logger.warning("Failed to reset name at 0x%X" % self.start_ea)
        elif ida_name.force_name(self.start_ea, new_name):
            self.name = idaapi.get_name(self.start_ea)
            if self.name != new_name:
                logger.info('IDA changed name "%s" to "%s"' % (new_name, self.name))
            return self.name
        else:
            logger.warning("Failed to rename at 0x%X" % self.start_ea)

    @property
    def api_calls(self) -> collections.Counter:
        """
        Returns counter containing API calls and the number of times they were called.
        """
        if self._api_calls:
            return self._api_calls

        api_calls = collections.Counter()
        for ea, func_ea in self.calls_from:
            func_name = ida_name.get_name(func_ea)
            if func_name:
                api_calls.update([func_name])

        self._api_calls = api_calls
        return self._api_calls

    @property
    def is_library(self) -> bool:
        """
        Is the function a library?
        """
        return bool(self.function_obj.flags & idc.FUNC_LIB)

    @property
    def calls_to(self) -> Iterable[int]:
        """Iterates addresses that call this function."""
        for ea in self.xrefs_to:
            insn = ida_ua.insn_t()
            ida_ua.decode_insn(insn, ea)
            if ida_idp.is_call_insn(insn):
                yield ea

    @property
    def calls_from(self) -> Iterable[Tuple[int, int]]:
        """Iterates call address and callee address of the calls within this function."""
        for ea in self.heads():
            insn = ida_ua.insn_t()
            ida_ua.decode_insn(insn, ea)
            if ida_idp.is_call_insn(insn):
                for xref in idautils.XrefsFrom(ea, idaapi.XREF_FAR):
                    func_ea = xref.to
                    if func_ea:
                        yield ea, func_ea

    @property
    def callers(self) -> Iterable["Function"]:
        """Iterates Function objects that call this function."""
        cache = set()
        for ea in self.calls_to:
            try:
                func = Function(ea)
            except AttributeError:
                continue
            if func.name not in cache:
                yield func
                cache.add(func.name)

    @property
    def callees(self) -> Iterable["Function"]:
        """Iterates Function objects that this function calls."""
        cache = set()
        for ea, func_ea in self.calls_from:
            try:
                func = Function(func_ea)
            except AttributeError:
                continue
            if func.name not in cache:
                yield func
                cache.add(func.name)

    @property
    def data(self) -> bytes:
        """Returns all the bytes contained in the function."""
        return ida_bytes.get_bytes(self.start_ea, self.end_ea - self.start_ea)



# Constants that will be used throughout the Windows API functions

from enum import IntEnum, IntFlag


MIN_HANDLE = 0x80               # Max the lowest handle reasonable
MAX_HANDLE = 0xFFFFFFFF - 1     # set the highest handle value

# File creation disposition
CREATE_NEW = 1
CREATE_ALWAYS = 2
OPEN_EXISTING = 3
OPEN_ALWAYS = 4
TRUNCATE_EXISTING = 5

# File access constants
GENERIC_READ = 0x80000000
GENERIC_WRITE = 0x40000000

# File share mode
FILE_SHARE_READ = 1
FILE_SHARE_WRITE = 2
FILE_SHARE_DELETE = 4

# Standard success stats
ERROR_SUCCESS = 0

# Registry view
KEY_WOW64_32KEY = 0x0200
KEY_WOW64_64KEY = 0x0100

# Registry disposition
REG_CREATED_NEW_KEY = 1
REG_OPENED_EXISTING_KEY = 2


class CSIDL(IntEnum):
    """"""
    CSIDL_PROGRAMS = 0x02
    CSIDL_PERSONAL = 0x05
    CSIDL_FAVORITES = 0x06
    CSIDL_STARTUP = 0x07
    CSIDL_RECENT = 0x08
    CSIDL_SENDTO = 0x09
    CSIDL_STARTMENU = 0x0b
    CSIDL_MYMUSIC = 0x0d
    CSIDL_MYVIDEO = 0x0e
    CSIDL_DESKTOPDIRECTORY = 0x10
    CSIDL_NETHOOD = 0x13
    CSIDL_FONTS = 0x14
    CSIDL_TEMPLATES = 0x15
    CSIDL_COMMON_STARTMENU = 0x16
    CSIDL_COMMON_PROGRAMS = 0x17
    CSIDL_COMMON_STARTUP = 0x18
    CSIDL_COMMON_DESKTOPDIRECTORY = 0x19
    CSIDL_APPDATA = 0x1a
    CSIDL_PRINTHOOD = 0x1b
    CSIDL_LOCAL_APPDATA = 0x1c
    CSIDL_COMMON_FAVORITES = 0x1f
    CSIDL_INTERNET_CACHE = 0x20
    CSIDL_COOKIES = 0x21
    CSIDL_HISTORY = 0x22
    CSIDL_COMMON_APPDATA = 0x23
    CSIDL_WINDOWS = 0x24
    CSIDL_SYSTEM = 0x25
    CSIDL_PROGRAM_FILES = 0x26
    CSIDL_MYPICTURES = 0x27
    CSIDL_PROFILE = 0x28
    CSIDL_SYSTEMX86 = 0x29
    CSIDL_PROGRAM_FILESX86 = 0x2a
    CSIDL_PROGRAM_FILES_COMMON = 0x2b
    CSIDL_PROGRAM_FILES_COMMONX86 = 0x2c
    CSIDL_COMMON_TEMPLATES = 0x2d
    CSIDL_COMMON_DOCUMENTS = 0x2e
    CSIDL_COMMON_ADMINTOOLS = 0x2f
    CSIDL_ADMINTOOLS = 0x30
    CSIDL_COMMON_MUSIC = 0x35
    CSIDL_COMMON_PICTURES = 0x36
    CSIDL_COMMON_VIDEO = 0x37
    CSIDL_CDBURN_AREA = 0x3b


class CSIDL_Flags(IntFlag):
    """CSIDL Flags"""
    CSIDL_FLAG_CREATE = 0x8000
    CSIDL_FLAG_DONT_UNEXPAND = 0x2000
    CSIDL_FLAG_DONT_VERIFY = 0x4000
    CSIDL_FLAG_NO_ALIAS = 0x1000
    CSIDL_FLAG_PER_USER_INIT = 0x0800
    CSIDL_FLAG_MASK = 0xff00


class RegistryKey(IntEnum):
    """Predefined keys in winreg.h"""
    HKEY_CLASSES_ROOT = 0x80000000
    HKEY_CURRENT_USER = 0x80000001
    HKEY_LOCAL_MACHINE = 0x80000002
    HKEY_USERS = 0x80000003
    HKEY_PERFORMANCE_DATA = 0x80000004
    HKEY_CURRENT_CONFIG = 0x80000005
    HKEY_DYN_DATA = 0x80000006
    HKEY_PERFORMANCE_TEXT = 0x80000050
    HKEY_PERFORMANCE_NLSTEXT = 0x80000060


class RegistryDataType(IntEnum):
    """Registry value data types in winreg.h"""
    REG_NONE = 0
    REG_SZ = 1
    REG_EXPAND_SZ = 2
    REG_BINARY = 3
    REG_DWORD = 4
    REG_DWORD_BIG_ENDIAN = 5
    REG_LINK = 6
    REG_MULTI_SZ = 7
    REG_QWORD = 11


class ServiceAccess(IntFlag):
    """Access Rights for a Service"""
    SERVICE_ALL_ACCESS = 0xF01FF
    SERVICE_CHANGE_CONFIG = 0x0002
    SERVICE_ENUMERATE_DEPENDENTS = 0x0008
    SERVICE_INTERROGATE = 0x0080
    SERVICE_PAUSE_CONTINUE = 0x0040
    SERVICE_QUERY_CONFIG = 0x0001
    SERVICE_QUERY_STATUS = 0x0004
    SERVICE_START = 0x0010
    SERVICE_STOP = 0x0020
    SERVICE_USER_DEFINED_CONTROL = 0x0100


class ServiceType(IntFlag):
    SERVICE_ADAPTER = 0x004
    SERVICE_FILE_SYSTEM_DRIVER = 0x002
    SERVICE_KERNEL_DRIVER = 0x001
    SERVICE_RECOGNIZER_DRIVER = 0x008
    SERVICE_WIN32_OWN_PROCESS = 0x010
    SERVICE_WIN32_SHARE_PROCESS = 0x020
    SERVICE_USER_OWN_PROCESS = 0x050
    SERVICE_USER_SHARE_PROCESS = 0x060
    SERVICE_INTERACTIVE_PROCESS = 0x100


class ServiceStart(IntEnum):
    SERVICE_AUTO_START = 2
    SERVICE_BOOT_START = 0
    SERVICE_DEMAND_START = 3
    SERVICE_DISABLED = 4
    SERVICE_SYSTEM_START = 1


class Visibility(IntEnum):
    """Enums for nCmdShow parameter."""
    SW_FORCEMINIMIZE = 11
    SW_HIDE = 0
    SW_MAXIMIZE = 3
    SW_MINIMIZE = 6
    SW_RESTORE = 9
    SW_SHOW = 5
    SW_SHOWDEFAULT = 10
    SW_SHOWMAXIMIZED = 3
    SW_SHOWMINIMIZED = 2
    SW_SHOWMINNOACTIVE = 7
    SW_SHOWNA = 8
    SW_SHOWNOACTIVATE = 4
    SW_SHOWNORMAL = 1
